from typing import AsyncGenerator, Literal, Sized

from nova.api import models
from nova.core import logger
from nova.core.gateway import NovaDevice
from nova.core.io import IOAccess
from nova.core.motion_group import MotionGroup
from nova.core.robot_cell import AbstractController, AbstractRobot, IODevice, ValueType


class Controller(Sized, AbstractController, NovaDevice, IODevice):
    """
    Represents a Nova controller, managing motion groups and IO interactions.
    """

    class Configuration(NovaDevice.Configuration):
        type: Literal["controller"] = "controller"
        id: str = "controller"
        cell_id: str
        controller_id: str

    def __init__(self, configuration: Configuration):
        super().__init__(configuration)
        self._activated_motion_group_ids: list[str] = []
        self._io_access = IOAccess(
            api_gateway=self._nova_api,
            cell=self.configuration.cell_id,
            controller_id=self.configuration.controller_id,
        )

    @property
    def controller_id(self) -> str:
        """Returns the unique controller ID."""
        return self.configuration.controller_id

    async def open(self):
        """Activates all motion groups."""
        motion_group_ids = await self.activated_motion_group_ids()
        self._activated_motion_group_ids = motion_group_ids
        logger.info(f"Found motion group {motion_group_ids}")
        return self

    async def close(self):
        # RPS-1174: when a motion group is deactivated, RAE closes all open connections
        #           this behaviour is not desired in some cases,
        #           so for now we will not deactivate for the user
        pass

    def __len__(self) -> int:
        return len(self._activated_motion_group_ids)

    def motion_group(self, motion_group_id: str) -> MotionGroup:
        """Returns motion group with specific id.

        Args:
            motion_group_id (str): The ID of the motion group.

        Returns:
            MotionGroup: A MotionGroup instance corresponding to the given ID.
        """
        return MotionGroup(
            api_gateway=self._nova_api,
            cell=self.configuration.cell_id,
            motion_group_id=motion_group_id,
        )

    def __getitem__(self, motion_group_id: int) -> MotionGroup:
        return self.motion_group(f"{motion_group_id}@{self.configuration.controller_id}")

    async def activated_motion_group_ids(self) -> list[str]:
        """Activates and retrieves the list of motion group IDs available on this controller.

        The system automatically activates all motion groups on the associated controller.

        Returns:
            list[str]: A list of activated motion group IDs (e.g., ["0@controller_id"]).
        """
        return await self._nova_api.activate_all_motion_groups(
            cell=self.configuration.cell_id, controller=self.configuration.controller_id
        )

    async def activated_motion_groups(self) -> list[MotionGroup]:
        """Retrieves a list of `MotionGroup` instances for all activated motion groups.

        Returns:
            list[MotionGroup]: All activated motion groups as `MotionGroup` objects.
        """
        motion_group_ids = await self.activated_motion_group_ids()
        return [self.motion_group(motion_group_id) for motion_group_id in motion_group_ids]

    def get_robots(self) -> dict[str, AbstractRobot]:
        """Retrieves a dictionary of motion group IDs to their corresponding robots.

        Note:
            This method interprets motion group IDs to create corresponding `MotionGroup`
            objects and returns them as `AbstractRobot` references.

        Returns:
            dict[str, AbstractRobot]: A mapping of motion group ID to `MotionGroup` instance.
        """
        return {
            motion_group_id: self.motion_group(motion_group_id)
            for motion_group_id in self._activated_motion_group_ids
        }

    async def read(self, key: str) -> ValueType:
        """Reads an IO value from the controller.

        Args:
            key (str): The IO key name.

        Returns:
            ValueType: The value associated with the specified IO key.
        """
        return await self._io_access.read(key)

    async def write(self, key: str, value: ValueType) -> None:
        """Writes an IO value to the controller.

        Args:
            key (str): The IO key name.
            value (ValueType): The value to write to the specified IO key.
        """
        return await self._io_access.write(key, value)

    async def stream_state(self, rate_msecs) -> AsyncGenerator[models.RobotControllerState, None]:
        async for state in self._nova_api.stream_robot_controller_state(
            cell=self.configuration.cell_id,
            controller_id=self.configuration.controller_id,
            response_rate=rate_msecs,
        ):
            yield state



from math import pi

from wandelbots_api_client.models.abb_controller import AbbController
from wandelbots_api_client.models.fanuc_controller import FanucController
from wandelbots_api_client.models.kuka_controller import KukaController
from wandelbots_api_client.models.universalrobots_controller import UniversalrobotsController
from wandelbots_api_client.models.virtual_controller import VirtualController
from wandelbots_api_client.models.yaskawa_controller import YaskawaController

from nova import api

MANUFACTURER_HOME_POSITIONS = {
    api.models.Manufacturer.ABB: [0.0, 0.0, 0.0, 0.0, pi / 2, 0.0, 0.0],
    api.models.Manufacturer.FANUC: [0.0, 0.0, 0.0, 0.0, -pi / 2, 0.0, 0.0],
    api.models.Manufacturer.YASKAWA: [0.0, 0.0, 0.0, 0.0, -pi / 2, 0.0, 0.0],
    api.models.Manufacturer.KUKA: [0.0, -pi / 2, pi / 2, 0.0, pi / 2, 0.0, 0.0],
    api.models.Manufacturer.UNIVERSALROBOTS: [0.0, -pi / 2, -pi / 2, -pi / 2, pi / 2, -pi / 2, 0.0],
}


def _build_controller(
    name: str,
    controller: AbbController
    | FanucController
    | KukaController
    | UniversalrobotsController
    | VirtualController
    | YaskawaController,
) -> api.models.RobotController:
    """Helper function to wrap a controller configuration in a RobotController object."""
    return api.models.RobotController(
        name=name, configuration=api.models.RobotControllerConfiguration(controller)
    )


def abb_controller(
    name: str, controller_ip: str, egm_server_ip: str, egm_server_port: int
) -> api.models.RobotController:
    """
    Create an ABB controller configuration for a physical robot.
    Args:
        name (str): The name of the controller.
        controller_ip (str): The IP address of the ABB robot.
        egm_server_ip (str): The IP address of the EGM server.
        egm_server_port (str): The port of the EGM server.
    """
    abb_config = AbbController(
        controllerIp=controller_ip,
        egmServer=api.models.AbbControllerEgmServer(ip=egm_server_ip, port=egm_server_port),
    )
    return _build_controller(name=name, controller=abb_config)


def universal_robots_controller(name: str, controller_ip: str) -> api.models.RobotController:
    """
    Create a Universal Robots controller configuration for a physical robot.
    Args:
        name (str): The name of the controller.
        controller_ip (str): The IP address of the Universal Robots robot.
    """
    universal_config = UniversalrobotsController(controllerIp=controller_ip)
    return _build_controller(name=name, controller=universal_config)


def kuka_controller(
    name: str, controller_ip: str, controller_port: int, rsi_server_ip: str, rsi_server_port: int
) -> api.models.RobotController:
    """
    Create a KUKA controller configuration for a physical robot.
    Args:
        name (str): The name of the controller.
        controller_ip (str): The IP address of the KUKA robot.
        controller_port (str): The port of the KUKA robot.
        rsi_server_ip (str): The IP address of the RSI server.
        rsi_server_port (str): The port of the RSI server.
    """
    kuka_config = KukaController(
        controllerIp=controller_ip,
        controllerPort=controller_port,
        rsiServer=api.models.KukaControllerRsiServer(ip=rsi_server_ip, port=rsi_server_port),
    )
    return _build_controller(name=name, controller=kuka_config)


def fanuc_controller(name: str, controller_ip: str) -> api.models.RobotController:
    """
    Create a FANUC controller configuration for a physical robot.
    Args:
        name (str): The name of the controller.
        controller_ip (str): The IP address of the FANUC robot.
    """
    fanuc_config = FanucController(controllerIp=controller_ip)
    return _build_controller(name=name, controller=fanuc_config)


def yaskawa_controller(name: str, controller_ip: str) -> api.models.RobotController:
    """
    Create a Yaskawa controller configuration for a physical robot.
    Args:
        name (str): The name of the controller.
        controller_ip (str): The IP address of the Yaskawa robot.
    """
    yaskawa_config = YaskawaController(controllerIp=controller_ip)
    return _build_controller(name=name, controller=yaskawa_config)


def virtual_controller(
    name: str,
    manufacturer: api.models.Manufacturer,
    type: api.models.VirtualControllerTypes | None = None,
    json: str | None = None,
    position: list[float] | str | None = None,
) -> api.models.RobotController:
    """
    Create a virtual controller configuration.
    Args:
        name (str): The name of the controller.
        manufacturer (api.models.Manufacturer): The manufacturer of the robot.
        type (api.models.VirtualControllerTypes | None): One of the available virtual controller types for this manufacturer.
        position: (list[float] | None): Initial joint position of the first motion group from the virtual robot controller.
        json (str | None): Additional data to save on controller.
    """
    if position is None:
        position = str(MANUFACTURER_HOME_POSITIONS.get(manufacturer, [0.0] * 7))

    if isinstance(position, list):
        position = str(position)

    virtual_config = api.models.VirtualController(
        manufacturer=manufacturer, type=type, json=json, position=position
    )
    return _build_controller(name=name, controller=virtual_config)



from nova.actions.motions import collision_free
from nova.core.motion_group import MotionGroup
from nova_rerun_bridge.benchmark.benchmark_base import BenchmarkStrategy, run_benchmark


class CollisionFreeStrategy(BenchmarkStrategy):
    name = "collision_free"

    async def plan(
        self,
        motion_group: MotionGroup,
        target,
        collision_scene,
        tcp,
        optimizer_setup,
        nova,
        start_joint_position,
    ):
        return await motion_group.plan(
            [collision_free(target=target, collision_scene=collision_scene)], tcp=tcp
        )


if __name__ == "__main__":
    import asyncio

    asyncio.run(run_benchmark(CollisionFreeStrategy()))



"""
Example of planning a collision free PTP motion. A sphere is placed in the robot's path and the robot uses collision free p2p to move around it.
"""

import asyncio

import numpy as np
import rerun as rr
from nova import MotionSettings
from nova.actions import cartesian_ptp, collision_free
from nova.api import models
from nova.core.exceptions import PlanTrajectoryFailed
from nova.core.nova import Nova
from nova.types import Pose
from nova_rerun_bridge import NovaRerunBridge
from wandelbots_api_client.models import (
    CoordinateSystem,
    RotationAngles,
    RotationAngleTypes,
    Vector3d,
)


async def build_collision_world(
    nova: Nova, cell_name: str, robot_setup: models.OptimizerSetup
) -> str:
    collision_api = nova._api_client.store_collision_components_api
    scene_api = nova._api_client.store_collision_scenes_api

    # define annoying obstacle
    sphere_collider = models.Collider(
        shape=models.ColliderShape(models.Sphere2(radius=100, shape_type="sphere")),
        pose=models.Pose2(position=[-100, -500, 200]),
    )
    await collision_api.store_collider(
        cell=cell_name, collider="annoying_obstacle", collider2=sphere_collider
    )

    # define TCP collider geometry
    tool_collider = models.Collider(
        shape=models.ColliderShape(
            models.Box2(size_x=100, size_y=100, size_z=100, shape_type="box", box_type="FULL")
        )
    )
    await collision_api.store_collision_tool(
        cell=cell_name, tool="tool_box", request_body={"tool_collider": tool_collider}
    )

    # define robot link geometries
    robot_link_colliders = await collision_api.get_default_link_chain(
        cell=cell_name, motion_group_model=robot_setup.motion_group_type
    )
    await collision_api.store_collision_link_chain(
        cell=cell_name, link_chain="robot_links", collider=robot_link_colliders
    )

    # assemble scene
    scene = models.CollisionScene(
        colliders={"annoying_obstacle": sphere_collider},
        motion_groups={
            robot_setup.motion_group_type: models.CollisionMotionGroup(
                tool={"tool_geometry": tool_collider}, link_chain=robot_link_colliders
            )
        },
    )
    scene_id = "collision_scene"
    await scene_api.store_collision_scene(
        cell_name, scene_id, models.CollisionSceneAssembly(scene=scene)
    )
    return scene_id


async def test():
    async with Nova() as nova, NovaRerunBridge(nova) as bridge:
        await bridge.setup_blueprint()

        cell = nova.cell()
        controller = await cell.ensure_virtual_robot_controller(
            "ur5",
            models.VirtualControllerTypes.UNIVERSALROBOTS_MINUS_UR5E,
            models.Manufacturer.UNIVERSALROBOTS,
        )

        await nova._api_client.virtual_robot_setup_api.set_virtual_robot_mounting(
            cell="cell",
            controller=controller.controller_id,
            id=0,
            coordinate_system=CoordinateSystem(
                coordinate_system="world",
                name="mounting",
                reference_uid="",
                position=Vector3d(x=0, y=0, z=0),
                rotation=RotationAngles(
                    angles=[0, 0, 0], type=RotationAngleTypes.EULER_ANGLES_EXTRINSIC_XYZ
                ),
            ),
        )

        # NC-1047
        await asyncio.sleep(5)

        # Connect to the controller and activate motion groups
        async with controller[0] as motion_group:
            await bridge.log_saftey_zones(motion_group)

            tcp = "Flange"

            robot_setup: models.OptimizerSetup = await motion_group._get_optimizer_setup(tcp=tcp)
            robot_setup.safety_setup.global_limits.tcp_velocity_limit = 200

            collision_scene_id = await build_collision_world(nova, "cell", robot_setup)

            await bridge.log_collision_scenes()

            # Use default planner to move to the right of the sphere
            home = await motion_group.tcp_pose(tcp)
            actions = [
                cartesian_ptp(home),
                cartesian_ptp(target=Pose((300, -400, 200, np.pi, 0, 0))),
            ]

            for action in actions:
                action.settings = MotionSettings(tcp_velocity_limit=200)

            try:
                joint_trajectory = await motion_group.plan(
                    actions, tcp, start_joint_position=(0, -np.pi / 2, np.pi / 2, 0, 0, 0)
                )
                await bridge.log_actions(actions)
                await bridge.log_trajectory(joint_trajectory, tcp, motion_group)
            except PlanTrajectoryFailed as e:
                await bridge.log_actions(actions)
                await bridge.log_trajectory(e.error.joint_trajectory, tcp, motion_group)
                await bridge.log_error_feedback(e.error.error_feedback)

            rr.log(
                "motion/target_", rr.Points3D([[-500, -400, 200]], radii=[10], colors=[(0, 255, 0)])
            )

            # Use default planner to move to the left of the sphere
            # -> this will collide
            # only plan don't move
            actions = [cartesian_ptp(target=Pose((-500, -400, 200, np.pi, 0, 0)))]

            for action in actions:
                action.settings = MotionSettings(tcp_velocity_limit=200)

            try:
                joint_trajectory_with_collision = await motion_group.plan(
                    actions, tcp, start_joint_position=joint_trajectory.joint_positions[-1].joints
                )
                await bridge.log_actions(actions)
                await bridge.log_trajectory(joint_trajectory_with_collision, tcp, motion_group)
            except PlanTrajectoryFailed as e:
                await bridge.log_actions(actions)
                await bridge.log_trajectory(e.error.joint_trajectory, tcp, motion_group)
                await bridge.log_error_feedback(e.error.error_feedback)

            # Plan collision free PTP motion around the sphere
            scene_api = nova._api_client.store_collision_scenes_api
            collision_scene = await scene_api.get_stored_collision_scene(
                cell="cell", scene=collision_scene_id
            )

            welding_actions = [
                collision_free(
                    target=Pose((-500, -400, 200, np.pi, 0, 0)),
                    collision_scene=collision_scene,
                    settings=MotionSettings(tcp_velocity_limit=30),
                )
            ]

            trajectory_plan_combined = await motion_group.plan(
                welding_actions,
                tcp=tcp,
                start_joint_position=joint_trajectory.joint_positions[-1].joints,
            )
            await bridge.log_actions(welding_actions)
            await bridge.log_trajectory(trajectory_plan_combined, tcp, motion_group)


if __name__ == "__main__":
    asyncio.run(test())



import asyncio
import json
import os

from numpy import pi

from nova import Controller, Nova
from nova.actions import cartesian_ptp, joint_ptp
from nova.api import models
from nova.types import MotionSettings, Pose
from nova_rerun_bridge import NovaRerunBridge
from nova_rerun_bridge.trajectory import TimingMode

"""
Example: Show an external axis with a robot moving in a cell.

Prerequisites:
- A cell with a robot with an positioner (external) axis: setup a yaskawa-ar1440 and import the yaskawa-ar1440-with-external-axis.yaml in the settings app
- Set mounting of 16@yaskawa to yaskawa-ar1440-16-mounting
"""


async def move_robot(controller: Controller, bridge: NovaRerunBridge):
    async with controller[0] as motion_group:
        await bridge.log_saftey_zones(motion_group)

        tcp_names = await motion_group.tcp_names()
        tcp = tcp_names[0]
        home_joints = await motion_group.joints()

        current_pose = await motion_group.tcp_pose(tcp)

        actions = [
            joint_ptp(home_joints),
            cartesian_ptp(current_pose @ Pose((0, 0, -100, 0, -pi / 2, 0))),
            cartesian_ptp(current_pose @ Pose((-500, 0, 0, 0, -pi / 2, 0))),
            joint_ptp(home_joints),
        ]

        for action in actions:
            action.settings = MotionSettings(tcp_velocity_limit=200)

        trajectory = await motion_group.plan(actions, tcp)
        await bridge.log_trajectory(trajectory, tcp, motion_group, timing_mode=TimingMode.SYNC)


async def move_positioner(controller: Controller, bridge: NovaRerunBridge):
    async with controller[16] as motion_group:
        actions = [
            joint_ptp((0, 0)),
            joint_ptp((pi / 4, pi / 4)),
            joint_ptp((-pi / 4, -pi / 4)),
            joint_ptp((0, 0)),
        ]

        trajectory = await motion_group.plan(actions, "")
        await bridge.log_trajectory(trajectory, "", motion_group, timing_mode=TimingMode.SYNC)


async def main():
    async with Nova() as nova, NovaRerunBridge(nova) as bridge:
        await bridge.setup_blueprint()
        cell = nova.cell()

        # Load JSON configuration
        json_path = os.path.join(
            os.path.dirname(__file__), "yaskawa-ar1440-with-external-axis.json"
        )
        with open(json_path) as f:
            var_json = json.load(f)

        controller = await cell._get_controller_instance("yaskawa")
        if controller is None:
            await nova._api_client.controller_api.add_robot_controller(
                cell=cell._cell_id,
                robot_controller=models.RobotController(
                    name="yaskawa",
                    configuration=models.RobotControllerConfiguration(
                        models.VirtualController(
                            type="yaskawa-ar1440",
                            manufacturer="yaskawa",
                            position="[0,0,0,0,0,0]",
                            json=json.dumps(var_json),
                        )
                    ),
                ),
                completion_timeout=60 * 2,
            )
        controller = await cell.controller("yaskawa")

        await asyncio.gather(
            move_robot(controller, bridge=bridge), move_positioner(controller, bridge=bridge)
        )


if __name__ == "__main__":
    asyncio.run(main())



import asyncio
import time
from dataclasses import dataclass
from typing import Any, Protocol

import numpy as np
from wandelbots_api_client import models
from wandelbots_api_client.models import (
    CoordinateSystem,
    RotationAngles,
    RotationAngleTypes,
    Vector3d,
)

from nova.core.motion_group import MotionGroup
from nova.core.nova import Nova
from nova.types import Pose
from nova.types.vector3d import Vector3d as Vector3d_nova
from nova_rerun_bridge import NovaRerunBridge
from nova_rerun_bridge.benchmark.datasets import motion_benchmaker_raw
from nova_rerun_bridge.benchmark.log_successful_planning import log_successful_planning
from nova_rerun_bridge.benchmark.robometrics_helper import (
    convert_position,
    create_box_collider,
    create_cylinder_collider,
    quaternion_to_angle_axis,
)


@dataclass
class NovaMetrics:
    """Metrics for benchmark results."""

    success: bool = False
    time: float = 0.0
    position_error: float = 0.0
    orientation_error: float = 0.0
    motion_time: float = 0.0


class BenchmarkStrategy(Protocol):
    """Protocol for benchmark strategies."""

    name: str

    async def plan(
        self,
        motion_group: MotionGroup,
        target: Pose,
        collision_scene: models.CollisionScene,
        tcp: str,
        optimizer_setup: models.OptimizerSetup,
        nova: Nova,
        start_joint_position: tuple[float, ...],
    ) -> Any: ...


async def setup_collision_scene(
    nova: Nova,
    obstacles: dict[str, Any],
    cell_name: str,
    motion_group_type: str,
    robot_setup: models.OptimizerSetup,
    scene_key: str,
) -> str:
    """Convert robometrics obstacles to Nova collision world format."""
    collision_api = nova._api_client.store_collision_components_api
    scene_api = nova._api_client.store_collision_scenes_api

    colliders = {}

    # Convert cuboid obstacles
    if "cuboid" in obstacles:
        for name, cube in obstacles["cuboid"].items():
            name, collider = create_box_collider(name, cube)
            colliders[name] = collider

    # Convert cylinder obstacles
    if "cylinder" in obstacles:
        for name, cylinder in obstacles["cylinder"].items():
            name, collider = create_cylinder_collider(name, cylinder)
            colliders[name] = collider

    # Define TCP collider geometry
    tool_collider = models.Collider(
        shape=models.ColliderShape(
            models.Box2(
                size_x=0.1,  # 10cm box for TCP
                size_y=0.1,
                size_z=0.1,
                shape_type="box",
                box_type="FULL",
            )
        )
    )
    await collision_api.store_collision_tool(
        cell=cell_name, tool="tool_box", request_body={"tool_collider": tool_collider}
    )

    # Define robot link geometries
    robot_link_colliders = await collision_api.get_default_link_chain(
        cell=cell_name, motion_group_model=robot_setup.motion_group_type
    )
    await collision_api.store_collision_link_chain(
        cell=cell_name, link_chain="robot_links", collider=robot_link_colliders
    )

    # Create and store collision scene with motion group
    scene = models.CollisionScene(
        colliders=colliders,
        motion_groups={
            motion_group_type: models.CollisionMotionGroup(
                tool={"tool_geometry": tool_collider}, link_chain=robot_link_colliders
            )
        },
    )
    scene_id = f"benchmark_scene_{scene_key}"
    await scene_api.store_collision_scene(
        cell_name, scene_id, models.CollisionSceneAssembly(scene=scene)
    )

    return scene_id


def print_separator(char="=", width=40):
    """Print a separator line."""
    print(char * width)


def print_statistics(results, name="Overall", show_separator=True):
    """Print statistics for a group of results with improved formatting."""
    if show_separator:
        print_separator("-")

    if not results:
        print(f"\n📊 {name} Statistics: No results collected")
        return

    success_rate = sum(r.success for r in results) / len(results)
    successful_results = [r for r in results if r.success]

    print(f"\n📊 {name} Statistics:")
    print(f"├── Total attempts: {len(results)}")
    print(
        f"├── Success rate:  {'🟢' if success_rate > 0.8 else '🟡' if success_rate > 0.5 else '🔴'} {success_rate:.2%}"
    )

    if successful_results:
        avg_time = sum(r.time for r in successful_results) / len(successful_results)
        avg_pos_error = sum(r.position_error for r in successful_results) / len(successful_results)
        avg_orient_error = sum(r.orientation_error for r in successful_results) / len(
            successful_results
        )
        avg_motion_time = sum(r.motion_time for r in successful_results) / len(successful_results)

        print("├── Averages for successful attempts:")
        print(f"│   ├── Planning time:     {avg_time:.3f}s")
        print(f"│   ├── Motion time:       {avg_motion_time:.3f}s")
        print(f"│   ├── Position error:    {avg_pos_error:.3f}m")
        print(f"│   └── Orientation error: {avg_orient_error:.3f}rad")

    if show_separator:
        print_separator("-")


def print_progressive_statistics(
    results,
    problem_results,
    current_key=None,
    total_problems=0,
    current_problem=0,
    path_to_rrd=None,
):
    """Print both current problem and overall statistics with progress indicator."""
    print("\033[2J\033[H")  # Clear screen and move cursor to top
    print_separator()
    print(f"🤖 Benchmark Progress: {current_problem}/{total_problems} problems")
    print_separator()

    # Print current problem statistics if available
    if current_key and current_key in problem_results:
        print_statistics(
            problem_results[current_key], f"Current Scene ({current_key})", show_separator=False
        )

    # Print overall progress
    print_statistics(results, "Overall Progress", show_separator=False)
    print_separator()
    if path_to_rrd:
        print(f"💾 Logged rerun scene: {path_to_rrd}")
        print_separator()


async def run_single_benchmark(strategy: BenchmarkStrategy):
    async with Nova() as nova, NovaRerunBridge(nova, spawn=False) as bridge:
        cell = nova.cell()

        controller = await cell.ensure_virtual_robot_controller(
            "ur10",
            models.VirtualControllerTypes.UNIVERSALROBOTS_MINUS_UR10E,
            models.Manufacturer.UNIVERSALROBOTS,
        )

        # move the robot up to avoid collision with the floor
        await nova._api_client.virtual_robot_setup_api.set_virtual_robot_mounting(
            cell="cell",
            controller=controller.controller_id,
            id=0,
            coordinate_system=CoordinateSystem(
                coordinate_system="world",
                name="mounting",
                reference_uid="",
                position=Vector3d(x=0, y=0, z=10),
                rotation=RotationAngles(
                    angles=[0, 0, 0], type=RotationAngleTypes.EULER_ANGLES_EXTRINSIC_XYZ
                ),
            ),
        )

        # NC-1047
        await asyncio.sleep(3)

        # Load benchmark datasets
        problems = motion_benchmaker_raw()

        # Store results per problem
        problem_results: dict[str, list[NovaMetrics]] = {}
        results = []

        total_problems = sum(len(probs) for probs in problems.values())
        current_problem = 0

        for key, scene_problems in problems.items():
            print(f"\nProcessing scene: {key}")
            problem_results[key] = []

            for i, problem in enumerate(scene_problems):
                current_problem += 1

                try:
                    async with controller[0] as motion_group:
                        tcp = "Flange"

                        # Get robot setup for collision scene
                        robot_setup: models.OptimizerSetup = (
                            await motion_group._get_optimizer_setup(tcp=tcp)
                        )

                        # Add collision objects from benchmark
                        collision_scene_id = await setup_collision_scene(
                            nova,
                            problem["obstacles"],
                            "cell",
                            robot_setup.motion_group_type,
                            robot_setup,
                            key,
                        )

                        scene_api = nova._api_client.store_collision_scenes_api
                        collision_scene = await scene_api.get_stored_collision_scene(
                            cell="cell", scene=collision_scene_id
                        )

                        # Convert start configuration to Pose object
                        position = convert_position(problem["goal_pose"]["position_xyz"])
                        orientation = quaternion_to_angle_axis(
                            problem["goal_pose"]["quaternion_wxyz"]
                        )

                        start_pose = Pose(
                            position=Vector3d_nova(x=position[0], y=position[1], z=position[2]),
                            orientation=Vector3d_nova(
                                x=orientation[0], y=orientation[1], z=orientation[2]
                            ),
                        )  # Combine position and orientation

                        start_time = time.time()

                        # Use the provided strategy
                        trajectory = await strategy.plan(
                            motion_group=motion_group,
                            target=start_pose,
                            collision_scene=collision_scene,
                            tcp=tcp,
                            optimizer_setup=robot_setup,
                            nova=nova,
                            start_joint_position=(0, -np.pi / 2, np.pi / 2, 0, 0, 0),
                        )

                        # Calculate metrics
                        end_time = time.time()
                        metrics = NovaMetrics(
                            success=True,
                            time=end_time - start_time,
                            position_error=0,  # not supported
                            orientation_error=0,  # not supported
                            motion_time=trajectory.times[-1] if trajectory else 0.0,
                        )
                        results.append(metrics)
                        problem_results[key].append(metrics)

                        path_to_rrd = await log_successful_planning(
                            bridge,
                            strategy.name,
                            collision_scene_id,
                            problem,
                            key,
                            i,
                            trajectory,
                            tcp,
                            motion_group,
                        )

                        print_progressive_statistics(
                            results,
                            problem_results,
                            key,
                            total_problems,
                            current_problem,
                            path_to_rrd,
                        )

                except Exception:
                    print(f"\nFailed planning: {key} - {i}")
                    failed_metrics = NovaMetrics()
                    results.append(failed_metrics)
                    problem_results[key].append(failed_metrics)

                    print_progressive_statistics(
                        results, problem_results, key, total_problems, current_problem
                    )

        return problem_results, results


async def run_benchmark(strategy: BenchmarkStrategy):
    """Run benchmark for a specific strategy."""
    print(f"\n🚀 Running benchmark for {strategy.name}")
    print_separator("=")

    problem_results, results = await run_single_benchmark(strategy)

    # Print results for this strategy
    print(f"\n📊 Results for {strategy.name}")
    print_separator("=")

    for key, prob_results in problem_results.items():
        print_statistics(prob_results, f"Scene: {key}")

    print_separator("=")
    print(f"📈 Overall Results for {strategy.name}")
    print_separator("=")
    print_statistics(results, f"Complete Benchmark - {strategy.name}")



import asyncio

import httpx
import pydantic

from nova.auth.auth_config import Auth0Config


class Auth0DeviceCodeInfo(pydantic.BaseModel):
    """
    Model to store device code information.

    Attributes:
        device_code (str): The device code.
        user_code (str): The user code.
        verification_uri (str): The verification URI.
        expires_in (int): The expiration time in seconds.
        interval (int): The interval time in seconds (default is 5).
    """

    device_code: str
    user_code: str
    verification_uri: str
    expires_in: int
    interval: int = pydantic.Field(default=5)


class Auth0TokenInfo(pydantic.BaseModel):
    """
    Model to store token information.

    Attributes:
        access_token (str): The access token.
        refresh_token (str, optional): The refresh token.
    """

    access_token: str
    refresh_token: str | None = None


class Auth0Parameters(pydantic.BaseModel):
    """
    Model to store Auth0 parameters.

    Attributes:
        auth0_domain (str): The Auth0 domain.
        auth0_client_id (str): The Auth0 client ID.
        auth0_audience (str): The Auth0 audience.
    """

    auth0_domain: str
    auth0_client_id: str
    auth0_audience: str


class Auth0DeviceAuthorization:
    """
    Class to handle Auth0 device authorization.

    Methods:
        __init__(auth0_domain: str, auth0_client_id: str, auth0_audience: str):
            Initializes the Auth0DeviceAuthorization instance with the given parameters.
        request_device_code():
            Requests a device code from Auth0.
        display_user_instructions():
            Displays instructions for the user to authenticate.
        poll_token_endpoint():
            Polls the token endpoint to obtain an access token.
        refresh_access_token(refresh_token: str):
            Refreshes the access token using the refresh token.
    """

    def __init__(self, auth0_config: Auth0Config | None = None):
        """
        Initialize with Auth0Config from env vars or passed config.

        Args:
            auth0_config: Optional Auth0Config object. If not provided,
                         will be created from environment variables.
        """
        try:
            self.params = auth0_config or Auth0Config.from_env()
            if not self.params.is_complete():
                raise ValueError("Auth0 configuration is incomplete")

            domain, client_id, audience = self.params.get_validated_config()
            self.auth0_domain = domain
            self.auth0_client_id = client_id
            self.auth0_audience = audience

        except ValueError as e:
            raise ValueError(f"Error: Auth0 configuration is invalid: {e}")

        self.headers = {"content-type": "application/x-www-form-urlencoded"}
        self.device_code_info: Auth0DeviceCodeInfo | None = None
        self.interval = 5
        self.attempts = 10

    def request_device_code(self):
        """
        Requests a device code from Auth0.

        Returns:
            Auth0DeviceCodeInfo: The device code information.

        Raises:
            Exception: If there is an error requesting the device code.
        """
        device_code_url = f"https://{self.auth0_domain}/oauth/device/code"
        data = {
            "client_id": self.auth0_client_id,
            "scope": "openid profile email",
            "audience": self.auth0_audience,
        }

        response = httpx.post(device_code_url, headers=self.headers, data=data)
        if response.status_code == 200:
            self.device_code_info = Auth0DeviceCodeInfo(**response.json())
            self.interval = self.device_code_info.interval
            return self.device_code_info
        raise Exception("Error requesting device code:", response.json())

    def get_device_code_info(self) -> Auth0DeviceCodeInfo | None:
        """
        Returns the device code information.

        Returns:
            Auth0DeviceCodeInfo | None: The device code information.
        """
        return self.device_code_info

    def display_user_instructions(self) -> None:
        """
        Displays instructions for the user to authenticate.

        Raises:
            Exception: If device code information is not available.
        """
        if self.device_code_info:
            verification_uri = f"{self.device_code_info.verification_uri}?user_code={self.device_code_info.user_code}"
            user_code = self.device_code_info.user_code
            print(
                f"Please visit {verification_uri} and validate the code {user_code} to authenticate."
            )
        else:
            raise Exception("Device code information is not available.")

    async def poll_token_endpoint(self):
        """
        Polls the token endpoint to obtain an access token.

        Returns:
            str: The access token.

        Raises:
            Exception: If there is an error polling the token endpoint.
        """
        if not self.device_code_info:
            raise Exception("Device code information is not available.")

        token_url = f"https://{self.auth0_domain}/oauth/token"
        token_data = {
            "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
            "device_code": self.device_code_info.device_code,
            "client_id": self.auth0_client_id,
        }

        async with httpx.AsyncClient() as client:
            while self.attempts > 0:
                self.attempts = self.attempts - 1
                token_response = await client.post(token_url, headers=self.headers, data=token_data)
                if token_response.status_code == 200:
                    # If the response status is 200, it means the access token is successfully obtained.
                    token_info = Auth0TokenInfo(**token_response.json())
                    self.refresh_token = token_info.refresh_token
                    return token_info.access_token
                if token_response.status_code == 400:
                    # If the response status is 400, check the error type.
                    error = token_response.json().get("error")
                    if error == "authorization_pending":
                        # If the error is 'authorization_pending', it means the user has not yet authorized.
                        await asyncio.sleep(self.interval)
                    elif error == "slow_down":
                        # If the error is 'slow_down', it means the server requests to slow down polling.
                        self.interval += 5
                        await asyncio.sleep(self.interval)
                elif token_response.status_code == 403:
                    # If the response status is 403, it means the request is forbidden, wait and retry.
                    await asyncio.sleep(self.interval)
                else:
                    # For other status codes, raise an exception with the error details.
                    raise Exception("Error:", token_response.status_code, token_response.json())
            raise Exception("Error: It was not able to authenticate. Please try again.")

    def refresh_access_token(self, refresh_token: str):
        """
        Refreshes the access token using the refresh token.

        Args:
            refresh_token (str): The refresh token.

        Returns:
            str: The new access token.

        Raises:
            Exception: If there is an error refreshing the access token.
        """
        if not refresh_token:
            raise Exception("Refresh token is not available.")

        token_url = f"https://{self.auth0_domain}/oauth/token"
        token_data = {
            "grant_type": "refresh_token",
            "client_id": self.auth0_client_id,
            "refresh_token": refresh_token,
        }

        response = httpx.post(token_url, headers=self.headers, data=token_data)

        if response.status_code == 200:
            token_info = Auth0TokenInfo(**response.json())
            return token_info.access_token
        raise Exception("Error refreshing access token:", response.json())



import asyncio

from nova.api import models
from nova.core.nova import Nova
from nova_rerun_bridge import NovaRerunBridge


async def build_collision_world(
    nova: Nova, cell_name: str, robot_setup: models.OptimizerSetup
) -> str:
    collision_api = nova._api_client.store_collision_components_api
    scene_api = nova._api_client.store_collision_scenes_api

    # Load all colliders from the JSON data
    colliders = {}

    # Box collider
    box_collider = models.Collider(
        shape=models.ColliderShape(
            models.Box2(size_x=100, size_y=50, size_z=200, shape_type="box", box_type="FULL")
        ),
        pose=models.Pose2(position=[0, 400, 0], orientation=[0, 0, 0]),
    )
    colliders["box"] = box_collider

    # Sphere collider
    sphere_collider = models.Collider(
        shape=models.ColliderShape(models.Sphere2(radius=30, shape_type="sphere")),
        pose=models.Pose2(position=[0, 200, 0]),
    )
    colliders["sphere"] = sphere_collider

    # Cylinder collider
    cylinder_collider = models.Collider(
        shape=models.ColliderShape(models.Cylinder2(radius=30, height=100, shape_type="cylinder")),
        pose=models.Pose2(position=[0, -600, 0]),
    )
    colliders["cylinder"] = cylinder_collider

    # Capsule collider
    capsule_collider = models.Collider(
        shape=models.ColliderShape(
            models.Capsule2(radius=30, cylinder_height=100, shape_type="capsule")
        ),
        pose=models.Pose2(position=[0, -400, 0]),
    )
    colliders["capsule"] = capsule_collider

    # Rectangular capsule collider
    rect_capsule_collider = models.Collider(
        shape=models.ColliderShape(
            models.RectangularCapsule2(
                radius=30,
                sphere_center_distance_x=100,
                sphere_center_distance_y=50,
                shape_type="rectangular_capsule",
            )
        ),
        pose=models.Pose2(position=[0, -200, 0]),
    )
    colliders["rectangular_capsule"] = rect_capsule_collider

    # Rectangle collider
    rectangle_collider = models.Collider(
        shape=models.ColliderShape(
            models.Rectangle2(size_x=30, size_y=100, shape_type="rectangle")
        ),
        pose=models.Pose2(position=[0, 0, 0]),
    )
    colliders["rectangle"] = rectangle_collider

    # Add rotated variants (x-axis 45 degrees)
    for name, collider in list(colliders.items()):
        rotated = models.Collider(
            shape=collider.shape,
            pose=models.Pose2(
                position=[
                    -300,
                    collider.pose.position[1] if collider.pose and collider.pose.position else 0,
                    collider.pose.position[2] if collider.pose and collider.pose.position else 0,
                ],
                orientation=[0.7853981633974484, 0, 0],  # 45 degrees in radians
            ),
        )
        colliders[f"{name}_rot_x_45"] = rotated

    # Add rotated variants (y-axis 45 degrees)
    for name, collider in list(colliders.items()):
        if "_rot_" not in name:  # Only rotate original objects
            rotated = models.Collider(
                shape=collider.shape,
                pose=models.Pose2(
                    position=[
                        300,
                        collider.pose.position[1]
                        if collider.pose and collider.pose.position
                        else 0,
                        collider.pose.position[2]
                        if collider.pose and collider.pose.position
                        else 0,
                    ],
                    orientation=[0, 0.7853981633974484, 0],  # 45 degrees in radians
                ),
            )
            colliders[f"{name}_rot_y_45"] = rotated

    # Store all colliders
    for name, collider in colliders.items():
        await collision_api.store_collider(cell=cell_name, collider=name, collider2=collider)

    # Define TCP collider geometry
    tool_collider = models.Collider(
        shape=models.ColliderShape(
            models.Box2(size_x=100, size_y=100, size_z=100, shape_type="box", box_type="FULL")
        )
    )
    await collision_api.store_collision_tool(
        cell=cell_name, tool="tool_box", request_body={"tool_collider": tool_collider}
    )

    # Define robot link geometries
    robot_link_colliders = await collision_api.get_default_link_chain(
        cell=cell_name, motion_group_model=robot_setup.motion_group_type
    )
    await collision_api.store_collision_link_chain(
        cell=cell_name, link_chain="robot_links", collider=robot_link_colliders
    )

    # Assemble scene with all colliders
    scene = models.CollisionScene(
        colliders=colliders,
        motion_groups={
            robot_setup.motion_group_type: models.CollisionMotionGroup(
                tool={"tool_geometry": tool_collider}, link_chain=robot_link_colliders
            )
        },
    )
    scene_id = "collision_scene"
    await scene_api.store_collision_scene(
        cell_name, scene_id, models.CollisionSceneAssembly(scene=scene)
    )
    return scene_id


async def test():
    async with Nova() as nova, NovaRerunBridge(nova) as bridge:
        await bridge.setup_blueprint()

        cell = nova.cell()
        controller = await cell.ensure_virtual_robot_controller(
            "ur5",
            models.VirtualControllerTypes.UNIVERSALROBOTS_MINUS_UR5E,
            models.Manufacturer.UNIVERSALROBOTS,
        )
        # Connect to the controller and activate motion groups
        async with controller[0] as motion_group:
            await bridge.log_saftey_zones(motion_group)

            tcp = "Flange"

            robot_setup: models.OptimizerSetup = await motion_group._get_optimizer_setup(tcp=tcp)

            await build_collision_world(nova, "cell", robot_setup)

            await bridge.log_collision_scenes()


if __name__ == "__main__":
    asyncio.run(test())



import rerun as rr

from nova.api import models
from nova_rerun_bridge.hull_visualizer import HullVisualizer


def log_safety_zones(motion_group: str, optimizer_config: models.OptimizerSetup):
    """
    Log hull outlines for the safety zones defined in the optimizer configuration.
    """
    if optimizer_config.safety_setup.safety_zones is None:
        return

    for zone in optimizer_config.safety_setup.safety_zones:
        geom = zone.geometry
        zone_id = zone.id
        entity_path = f"{motion_group}/safety_zones/zone_{zone_id}"

        if geom.compound is not None:
            child_geoms = geom.compound.child_geometries
            polygons = HullVisualizer.compute_hull_outlines_from_geometries(child_geoms)
        elif geom.convex_hull is not None:

            class ChildWrapper:
                def __init__(self, convex_hull):
                    self.convex_hull = convex_hull

            c = ChildWrapper(geom.convex_hull)
            c.convex_hull = geom.convex_hull
            polygons = HullVisualizer.compute_hull_outlines_from_geometries([c])
        else:
            polygons = []

        # Log polygons as wireframe outlines
        if polygons:
            line_segments = [p.tolist() for p in polygons]
            rr.log(
                entity_path,
                rr.LineStrips3D(
                    line_segments, radii=rr.Radius.ui_points(0.75), colors=[[221, 193, 193, 255]]
                ),
                static=True,
            )



import asyncio

from nova import MotionSettings
from nova.actions import cartesian_ptp, linear
from nova.api import models
from nova.core.exceptions import PlanTrajectoryFailed
from nova.core.nova import Nova
from nova.types import Pose
from nova_rerun_bridge import NovaRerunBridge


async def test():
    async with Nova() as nova, NovaRerunBridge(nova) as bridge:
        await bridge.setup_blueprint()

        cell = nova.cell()
        controller = await cell.ensure_virtual_robot_controller(
            "ur10",
            models.VirtualControllerTypes.UNIVERSALROBOTS_MINUS_UR10E,
            models.Manufacturer.UNIVERSALROBOTS,
        )

        # Connect to the controller and activate motion groups
        async with controller[0] as motion_group:
            await bridge.log_saftey_zones(motion_group)

            tcp = "Flange"

            home = await motion_group.tcp_pose(tcp)

            actions = [
                action
                for _ in range(3)
                for action in [
                    cartesian_ptp(home),
                    linear(target=Pose((50, 20, 30, 0, 0, 0)) @ home),
                    linear(target=Pose((100, 20, 30, 0, 0, 0)) @ home),
                    linear(target=Pose((50, 20, 30, 0, 0, 0)) @ home),
                    cartesian_ptp(home),
                ]
            ]

            for action in actions:
                action.settings = MotionSettings(tcp_velocity_limit=200)

            try:
                joint_trajectory = await motion_group.plan(actions, tcp)
                await bridge.log_actions(actions)
                await bridge.log_trajectory(joint_trajectory, tcp, motion_group)
            except PlanTrajectoryFailed as e:
                await bridge.log_actions(actions)
                await bridge.log_trajectory(e.error.joint_trajectory, tcp, motion_group)
                await bridge.log_error_feedback(e.error.error_feedback)
                return


if __name__ == "__main__":
    asyncio.run(test())



"""
This example demonstrates how to stream the robot state into the timeline "time".
This can be used to record the actual motions of the robot.
"""

import asyncio
import signal
from contextlib import asynccontextmanager

from nova import MotionSettings
from nova.actions import cartesian_ptp, joint_ptp
from nova.api import models
from nova.core.nova import Nova
from nova.types import Pose
from nova_rerun_bridge import NovaRerunBridge


@asynccontextmanager
async def handle_shutdown():
    stop = asyncio.Future()

    def signal_handler():
        if not stop.done():
            stop.set_result(None)

    loop = asyncio.get_event_loop()
    loop.add_signal_handler(signal.SIGINT, signal_handler)
    try:
        yield stop
    finally:
        loop.remove_signal_handler(signal.SIGINT)


async def test():
    async with Nova() as nova, NovaRerunBridge(nova) as bridge:
        await bridge.setup_blueprint()

        cell = nova.cell()
        controller = await cell.ensure_virtual_robot_controller(
            "ur10",
            models.VirtualControllerTypes.UNIVERSALROBOTS_MINUS_UR10E,
            models.Manufacturer.UNIVERSALROBOTS,
        )

        # Connect to the controller and activate motion groups
        async with controller[0] as motion_group:
            # this streams the current robot state into the timeline "time"
            # use this to record actual motions of the robot
            await bridge.start_streaming(motion_group)

            # In addition to log the robot state you can still log other data
            # like trajectories into the "time_interval_0.016" timeline
            await bridge.log_saftey_zones(motion_group)

            home_joints = await motion_group.joints()
            tcp_names = await motion_group.tcp_names()
            tcp = tcp_names[0]

            # Get current TCP pose and offset it slightly along the x-axis
            current_pose = await motion_group.tcp_pose(tcp)
            target_pose = current_pose @ Pose((1, 0, 0, 0, 0, 0))

            actions = [
                joint_ptp(home_joints),
                cartesian_ptp(target_pose),
                joint_ptp(home_joints),
                cartesian_ptp(target_pose @ [100, 0, 0, 0, 0, 0]),
                joint_ptp(home_joints),
                cartesian_ptp(target_pose @ (100, 100, 0, 0, 0, 0)),
                joint_ptp(home_joints),
                cartesian_ptp(target_pose @ Pose((0, 100, 0, 0, 0, 0))),
                joint_ptp(home_joints),
            ]

            # you can update the settings of the action
            for action in actions:
                action.settings = MotionSettings(tcp_velocity_limit=200)

            joint_trajectory = await motion_group.plan(actions, tcp)

            await bridge.log_actions(actions)
            await bridge.log_trajectory(joint_trajectory, tcp, motion_group)

            # Keep streaming until Ctrl+C
            async with handle_shutdown() as stop:
                try:
                    await stop
                except asyncio.CancelledError:
                    pass
                finally:
                    await bridge.stop_streaming()


if __name__ == "__main__":
    asyncio.run(test())



from typing import Any

import numpy as np
from scipy.spatial.transform import Rotation
from wandelbots_api_client import models


def m_to_mm(value: float) -> float:
    """Convert centimeters to millimeters."""
    return value * 1000.0


def convert_position(position: list[float]) -> list[float]:
    """Convert position coordinates from m to mm."""
    return [m_to_mm(p) for p in position]


def quaternion_to_angle_axis(quaternion: list[float]) -> list[float]:
    """Convert quaternion [w, x, y, z] to angle-axis [rx, ry, rz].

    Args:
        quaternion: Quaternion in [w, x, y, z] format as used by robometrics
    Returns:
        Angle-axis representation [rx, ry, rz] in radians
    """
    # Convert [w,x,y,z] to [x,y,z,w] for scipy
    w, x, y, z = quaternion
    rot = Rotation.from_quat([x, y, z, w])
    return rot.as_rotvec()


def convert_pose_quaternion(pose: list[float]) -> tuple[list[float], list[float]]:
    """Convert pose from robometrics format to Nova format.

    The pose transformation order is:
    1. Convert position to mm
    2. Calculate orientation (quaternion to angle-axis)

    Args:
        pose: [x, y, z, w, x, y, z] list containing position in meters
             and quaternion in [w,x,y,z] format
    Returns:
        tuple of (position_mm, orientation_rad)
    """
    # First convert position from meters to mm
    position = convert_position(pose[:3])

    # Then handle orientation - quaternion already in [w,x,y,z] format
    angles = quaternion_to_angle_axis(pose[3:])

    return position, angles


def create_box_collider(name: str, cube: dict[str, Any]) -> tuple[str, models.Collider]:
    """Create a box collider with mm dimensions using ConvexHull2."""
    position, orientation = convert_pose_quaternion(cube["pose"])
    dims = convert_position(cube["dims"])

    # Create box vertices (8 corners)
    half_x, half_y, half_z = [d / 2 for d in dims]
    vertices = [
        [-half_x, -half_y, -half_z],
        [-half_x, -half_y, half_z],
        [-half_x, half_y, -half_z],
        [-half_x, half_y, half_z],
        [half_x, -half_y, -half_z],
        [half_x, -half_y, half_z],
        [half_x, half_y, -half_z],
        [half_x, half_y, half_z],
    ]

    # Apply rotation to vertices first
    rot = Rotation.from_rotvec(np.array(orientation))
    rotated_vertices = [rot.apply(v) for v in vertices]

    return name, models.Collider(
        shape=models.ColliderShape(
            models.ConvexHull2(vertices=rotated_vertices, shape_type="convex_hull")
        ),
        pose=models.Pose2(
            position=position,
            orientation=[0, 0, 0],  # Orientation already applied to vertices
        ),
    )


def create_cylinder_collider(name: str, cylinder: dict[str, Any]) -> tuple[str, models.Collider]:
    """Create a cylinder collider with mm dimensions using ConvexHull2."""
    position, orientation = convert_pose_quaternion(cylinder["pose"])
    radius = m_to_mm(cylinder["radius"])
    height = m_to_mm(cylinder["height"])

    # Create cylinder vertices (discretized circle at top and bottom)
    num_points = 16  # Number of points to approximate the circular cross-section
    angles = np.linspace(0, 2 * np.pi, num_points, endpoint=False)

    # Create points for top and bottom circles
    top_points = [[radius * np.cos(a), radius * np.sin(a), height / 2] for a in angles]
    bottom_points = [[radius * np.cos(a), radius * np.sin(a), -height / 2] for a in angles]
    vertices = top_points + bottom_points

    # Apply rotation to vertices first
    rot = Rotation.from_rotvec(np.array(orientation))
    rotated_vertices = [rot.apply(v) for v in vertices]

    return name, models.Collider(
        shape=models.ColliderShape(
            models.ConvexHull2(vertices=rotated_vertices, shape_type="convex_hull")
        ),
        pose=models.Pose2(
            position=position,
            orientation=[0, 0, 0],  # Orientation already applied to vertices
        ),
    )



import json

import wandelbots_api_client as wb


class ControllerNotFound(Exception):
    def __init__(self, controller: str):
        super().__init__(f"Controller {controller} not found.")


class PlanTrajectoryFailed(Exception):
    def __init__(self, error: wb.models.PlanTrajectoryFailedResponse, motion_group_id: str):
        """
        Create a PlanTrajectoryFailed exception.

        Args:
            error:           The failure response.
            motion_group_id: The ID of the motion group that caused the exception, e.g. `0@controller`
        """
        self._error = error
        self._motion_group_id = motion_group_id
        super().__init__(
            f"Plan trajectory on {motion_group_id} failed: {json.dumps(error.to_dict(), indent=2)}"
        )

    def to_pretty_string(self) -> str:
        """Give a more lightweight representation of the error, omitting some gritty details."""
        error_dict = self._error.to_dict()
        del error_dict["joint_trajectory"]
        return (
            f"Plan trajectory on {self._motion_group_id} failed: {json.dumps(error_dict, indent=2)}"
        )

    @property
    def error(self) -> wb.models.PlanTrajectoryFailedResponse:
        """Return the original PlanTrajectoryFailedResponse object."""
        return self._error


class InitMovementFailed(Exception):
    def __init__(self, error: wb.models.InitializeMovementResponseInitResponse):
        self._error = error
        super().__init__(f"Initial movement failed: {json.dumps(error.to_dict(), indent=2)}")

    @property
    def error(self) -> wb.models.InitializeMovementResponseInitResponse:
        """Return the original InitializeMovementResponseInitResponse object."""
        return self._error


class LoadPlanFailed(Exception):
    def __init__(self, error: wb.models.PlanSuccessfulResponse):
        self._error = error
        super().__init__(f"Load plan failed: {json.dumps(error.to_dict(), indent=2)}")

    @property
    def error(self) -> wb.models.PlanSuccessfulResponse:
        """Return the original PlanSuccessfulResponse object."""
        return self._error


class InconsistentCollisionScenes(Exception):
    """Raised when actions have different collision scenes."""

    def __init__(self, message: str):
        self._message = message
        super().__init__(message)

    @property
    def message(self) -> str:
        """Return the error message."""
        return self._message



from typing import AsyncIterator, Callable

from nova import api
from nova.types.motion_settings import MotionSettings
from nova.types.pose import Pose
from nova.types.state import MotionState, RobotState
from nova.types.vector3d import Vector3d

LoadPlanResponse = api.models.PlanSuccessfulResponse
InitialMovementStream = AsyncIterator[api.models.StreamMoveResponse]
InitialMovementConsumer = Callable[[api.models.StreamMoveResponse], None]
MovementResponse = api.models.ExecuteTrajectoryResponse | api.models.StreamMoveResponse
ExecuteTrajectoryRequestStream = AsyncIterator[api.models.ExecuteTrajectoryRequest]
ExecuteTrajectoryResponseStream = AsyncIterator[api.models.ExecuteTrajectoryResponse]
MovementControllerFunction = Callable[
    [ExecuteTrajectoryResponseStream], ExecuteTrajectoryRequestStream
]

__all__ = [
    "Vector3d",
    "Pose",
    "CollisionScene",
    "LoadPlanResponse",
    "InitialMovementStream",
    "InitialMovementConsumer",
    "MovementResponse",
    "MotionState",
    "RobotState",
    "MotionSettings",
    "ExecuteTrajectoryRequestStream",
    "ExecuteTrajectoryResponseStream",
    "MovementControllerFunction",
]



import logging
import sys

from decouple import config

LOG_LEVEL: str = config("LOG_LEVEL", default="INFO").upper()
LOG_FORMAT: str = config("LOG_FORMAT", default="%(asctime)s [%(levelname)s] %(name)s: %(message)s")
LOG_DATETIME_FORMAT: str = config("LOG_DATETIME_FORMAT", default="%Y-%m-%d %H:%M:%S")
LOGGER_NAME: str = config("LOGGER_NAME", default="wandelbots-nova")

# Setting up the underlying logger
formatter: logging.Formatter = logging.Formatter(LOG_FORMAT, datefmt=LOG_DATETIME_FORMAT)
handler: logging.StreamHandler = logging.StreamHandler(sys.stdout)
handler.setLevel(LOG_LEVEL)
handler.setFormatter(formatter)

logger: logging.Logger = logging.getLogger(LOGGER_NAME)
logger.setLevel(LOG_LEVEL)
logger.addHandler(handler)



from functools import singledispatch
from typing import Any, Callable

import wandelbots_api_client as wb

from nova.actions import MovementControllerContext
from nova.core import logger
from nova.core.exceptions import InitMovementFailed
from nova.types import (
    ExecuteTrajectoryRequestStream,
    ExecuteTrajectoryResponseStream,
    MotionState,
    MovementControllerFunction,
    Pose,
    RobotState,
)


@singledispatch
def movement_to_motion_state(movement: Any) -> MotionState:
    raise NotImplementedError(f"Unsupported movement type: {type(movement)}")


@movement_to_motion_state.register
def _(movement: wb.models.Movement) -> MotionState:
    """Convert a wb.models.Movement to a MotionState."""
    if (
        movement.movement.state is None
        or movement.movement.current_location is None
        or len(movement.movement.state.motion_groups) == 0
    ):
        assert False, "This should not happen"  # depending on NC-1105

    # TODO: in which cases do we have more than one motion group here?
    motion_group = movement.movement.state.motion_groups[0]
    return motion_group_state_to_motion_state(
        motion_group, float(movement.movement.current_location)
    )


@movement_to_motion_state.register
def _(movement: wb.models.StreamMoveResponse) -> MotionState:
    """Convert a wb.models.Movement to a MotionState."""
    if (
        movement.move_response is None
        or movement.state is None
        or movement.move_response.current_location_on_trajectory is None
        or len(movement.state.motion_groups) == 0
    ):
        assert False, "This should not happen"  # depending on NC-1105

    # TODO: in which cases do we have more than one motion group here?
    motion_group = movement.state.motion_groups[0]
    return motion_group_state_to_motion_state(
        motion_group, float(movement.move_response.current_location_on_trajectory)
    )


def motion_group_state_to_motion_state(
    motion_group_state: wb.models.MotionGroupState, path_parameter: float
) -> MotionState:
    tcp_pose = Pose(motion_group_state.tcp_pose)
    joints = (
        tuple(motion_group_state.joint_current.joints) if motion_group_state.joint_current else None
    )
    return MotionState(
        motion_group_id=motion_group_state.motion_group,
        path_parameter=path_parameter,
        state=RobotState(pose=tcp_pose, joints=joints),
    )


def move_forward(context: MovementControllerContext) -> MovementControllerFunction:
    """
    movement_controller is an async function that yields requests to the server.
    If a movement_consumer is provided, we'll asend() each wb.models.MovementMovement to it,
    letting it produce MotionState objects.
    """

    async def movement_controller(
        response_stream: ExecuteTrajectoryResponseStream,
    ) -> ExecuteTrajectoryRequestStream:
        # The first request is to initialize the movement
        yield wb.models.InitializeMovementRequest(trajectory=context.motion_id, initial_location=0)  # type: ignore

        # then we get the response
        initialize_movement_response = await anext(response_stream)
        if isinstance(
            initialize_movement_response.actual_instance, wb.models.InitializeMovementResponse
        ):
            r1 = initialize_movement_response.actual_instance
            if not r1.init_response.succeeded:
                raise InitMovementFailed(r1.init_response)

        # The second request is to start the movement
        set_io_list = context.combined_actions.to_set_io()
        yield wb.models.StartMovementRequest(
            set_ios=set_io_list, start_on_io=None, pause_on_io=None
        )  # type: ignore

        # then we wait until the movement is finished
        async for execute_trajectory_response in response_stream:
            instance = execute_trajectory_response.actual_instance
            # Stop when standstill indicates motion ended
            if isinstance(instance, wb.models.Standstill):
                if instance.standstill.reason == wb.models.StandstillReason.REASON_MOTION_ENDED:
                    return

    return movement_controller


def speed_up(
    context: MovementControllerContext, on_movement: Callable[[MotionState | None], None]
) -> MovementControllerFunction:
    async def movement_controller(
        response_stream: ExecuteTrajectoryResponseStream,
    ) -> ExecuteTrajectoryRequestStream:
        # The first request is to initialize the movement
        yield wb.models.InitializeMovementRequest(trajectory=context.motion_id, initial_location=0)  # type: ignore

        # then we get the response
        initialize_movement_response = await anext(response_stream)
        if isinstance(
            initialize_movement_response.actual_instance, wb.models.InitializeMovementResponse
        ):
            r1 = initialize_movement_response.actual_instance
            if not r1.init_response.succeeded:
                raise InitMovementFailed(r1.init_response)

        # The second request is to start the movement
        set_io_list = context.combined_actions.to_set_io()
        yield wb.models.StartMovementRequest(
            set_ios=set_io_list, start_on_io=None, pause_on_io=None
        )  # type: ignore

        counter = 0
        latest_speed = 10
        # then we wait until the movement is finished
        async for execute_trajectory_response in response_stream:
            counter += 1
            instance = execute_trajectory_response.actual_instance
            # Send the current location to the consume
            if isinstance(instance, wb.models.Movement):
                motion_state = movement_to_motion_state(instance)
                if motion_state:
                    on_movement(motion_state)

            # Terminate the generator
            if isinstance(instance, wb.models.Standstill):
                if instance.standstill.reason == wb.models.StandstillReason.REASON_MOTION_ENDED:
                    on_movement(None)
                    return

            if isinstance(instance, wb.models.PlaybackSpeedResponse):
                playback_speed = instance.playback_speed_response
                logger.info(f"Current playback speed: {playback_speed}")

            if counter % 10 == 0:
                yield wb.models.ExecuteTrajectoryRequest(
                    wb.models.PlaybackSpeedRequest(playback_speed_in_percent=latest_speed)
                )
                counter = 0
                latest_speed += 5
                if latest_speed > 100:
                    latest_speed = 100

    return movement_controller



from abc import ABC, abstractmethod

import pydantic


class Action(pydantic.BaseModel, ABC):
    @abstractmethod
    @pydantic.model_serializer
    def serialize_model(self):
        """Serialize the model to a dictionary"""

    @abstractmethod
    def is_motion(self) -> bool:
        """Return whether the action is a motion"""



"""
This example demonstrates a handover between four different robots brands.
Each robot picks up a cube and passes it to the next robot in the sequence.

All robots are operating in the world coordinate system.
"""

import asyncio
from dataclasses import dataclass
from math import pi

import numpy as np
from wandelbots_api_client.models import (
    CoordinateSystem,
    RotationAngles,
    RotationAngleTypes,
    Vector3d,
)

from nova import Controller, Nova
from nova.actions import Action, cartesian_ptp, joint_ptp
from nova.api import models
from nova.core.exceptions import PlanTrajectoryFailed
from nova_rerun_bridge import NovaRerunBridge
from nova_rerun_bridge.trajectory import TimingMode


@dataclass
class RobotPosition:
    """Represents a robot's position and associated poses"""

    mounting: Vector3d
    rotation: RotationAngles  # Added rotation field
    cube_position: tuple[float, float, float, float, float, float]
    handover_position: tuple[float, float, float, float, float, float]
    home_position: tuple[float, ...]
    motion_group_id: int


ROBOT_POSITIONS = {
    "FANUC": RobotPosition(
        mounting=Vector3d(x=600, y=0, z=350),
        rotation=RotationAngles(
            angles=[0, 0, 0],  # Facing +X
            type=RotationAngleTypes.EULER_ANGLES_EXTRINSIC_XYZ,
        ),
        cube_position=(1000, 0, 100, pi, 0, 0),  # TCP down
        # Halfway between FANUC and KUKA
        handover_position=(300, 300, 400, pi, 0, 0),  # TCP down for handover
        home_position=(0.0, 0.0, 0.0, 0.0, -pi / 2, 0.0),
        motion_group_id=1,
    ),
    "KUKA": RobotPosition(
        mounting=Vector3d(x=0, y=600, z=0),
        rotation=RotationAngles(
            angles=[0, 0, pi / 2],  # Facing +Y
            type=RotationAngleTypes.EULER_ANGLES_EXTRINSIC_XYZ,
        ),
        cube_position=(0, 1000, 100, pi, 0, 0),  # TCP down
        # Halfway between KUKA and YASKAWA
        handover_position=(-300, 300, 400, pi, 0, 0),  # TCP down for handover
        home_position=(0.0, -pi / 2, pi / 2, 0.0, pi / 2, 0.0),
        motion_group_id=0,
    ),
    "YASKAWA": RobotPosition(
        mounting=Vector3d(x=-600, y=0, z=350),
        rotation=RotationAngles(
            angles=[0, 0, pi],  # Facing -X
            type=RotationAngleTypes.EULER_ANGLES_EXTRINSIC_XYZ,
        ),
        cube_position=(-1000, 0, 100, pi, 0, 0),  # TCP down
        # Halfway between YASKAWA and ABB
        handover_position=(-300, -300, 400, pi, 0, 0),  # TCP down for handover
        home_position=(0.0, 0.0, 0.0, 0.0, -pi / 2, 0.0),
        motion_group_id=0,
    ),
    "ABB": RobotPosition(
        mounting=Vector3d(x=0, y=-600, z=0),
        rotation=RotationAngles(
            angles=[0, 0, -pi / 2],  # Facing -Y
            type=RotationAngleTypes.EULER_ANGLES_EXTRINSIC_XYZ,
        ),
        cube_position=(0, -1000, 100, pi, 0, 0),  # TCP down
        # Halfway between ABB and FANUC
        handover_position=(300, -300, 400, pi, 0, 0),  # TCP down for handover
        home_position=(0.0, 0.0, 0.0, 0.0, pi / 2, 0.0),
        motion_group_id=0,
    ),
}


async def pick_and_pass_cube(
    controller: Controller,
    bridge: NovaRerunBridge,
    pos_config: RobotPosition,
    action: str,
    motion_group_id: int = 0,
    sync: bool = False,
):
    """Handle cube picking and passing for a single robot"""
    async with controller[motion_group_id] as motion_group:
        await bridge.log_saftey_zones(motion_group)
        tcp = (await motion_group.tcp_names())[0]

        # Calculate handover position with orientation
        handover_pos = list(pos_config.handover_position)
        if action in ["pickup_and_handover", "go_to_handover"]:
            handover_pos[3:] = calculate_handover_orientation(
                pos_config.mounting, tuple(handover_pos), is_receiver=(action == "go_to_handover")
            )
        handover_target: tuple[float, float, float, float, float, float] = (
            handover_pos[0],
            handover_pos[1],
            handover_pos[2],
            handover_pos[3],
            handover_pos[4],
            handover_pos[5],
        )

        # Define actions based on robot state
        actions_map: dict[str, list[Action]] = {
            "pickup_and_handover": [
                cartesian_ptp(pos_config.cube_position),
                cartesian_ptp(handover_target),
            ],
            "go_to_handover": [cartesian_ptp(handover_target)],
            "go_home_with_cube": [joint_ptp(pos_config.home_position)],
            "place_cube": [
                cartesian_ptp(pos_config.cube_position),
                joint_ptp(pos_config.home_position),
            ],
        }

        try:
            joint_trajectory = await motion_group.plan(actions_map[action], tcp)
            await bridge.log_actions(actions_map[action])
            await bridge.log_trajectory(
                joint_trajectory,
                tcp,
                motion_group,
                timing_mode=TimingMode.SYNC if sync else TimingMode.CONTINUE,
            )
            await motion_group.execute(joint_trajectory, tcp, actions_map[action])
        except PlanTrajectoryFailed as e:
            await bridge.log_actions(actions_map[action], show_connection=True)
            if e.error.joint_trajectory:
                await bridge.log_trajectory(e.error.joint_trajectory, tcp, motion_group)
            await bridge.log_error_feedback(e.error.error_feedback)


async def move_to_initial_positions(
    robots: dict[str, Controller], bridge: NovaRerunBridge, positions: dict[str, RobotPosition]
) -> None:
    """Move all robots to their initial pickup positions"""
    tasks = []
    for robot_name, pos_config in positions.items():

        async def move_robot(robot_name=robot_name, pos_config=pos_config):
            async with robots[robot_name][pos_config.motion_group_id] as motion_group:
                tcp_names = await motion_group.tcp_names()
                tcp = tcp_names[0]

                # Move to home, then to cube position
                actions: list[Action] = [joint_ptp(pos_config.home_position)]

                try:
                    joint_trajectory = await motion_group.plan(actions, tcp)
                    await bridge.log_actions(actions)
                    await bridge.log_trajectory(
                        joint_trajectory, tcp, motion_group, timing_mode=TimingMode.SYNC
                    )
                    await motion_group.execute(joint_trajectory, tcp, actions=actions)
                except PlanTrajectoryFailed as e:
                    await bridge.log_actions(actions, show_connection=True)
                    if e.error.joint_trajectory:
                        await bridge.log_trajectory(e.error.joint_trajectory, tcp, motion_group)
                    await bridge.log_error_feedback(e.error.error_feedback)

        tasks.append(move_robot())

    await asyncio.gather(*tasks)


def calculate_handover_orientation(
    base_pos: Vector3d, handover_pos: tuple[float, ...], is_receiver: bool = False
) -> tuple[float, float, float]:
    """Calculate TCP orientation for handover using axis-angle representation"""
    direction = np.array(
        [
            handover_pos[0] - base_pos.x,
            handover_pos[1] - base_pos.y,
            0,  # Ignore Z for horizontal orientation
        ]
    )
    direction = direction / np.linalg.norm(direction)
    if not is_receiver:
        direction = -direction

    # Cross product with downward vector to get rotation axis
    rotation_axis = np.cross([0, 0, -1], direction)
    angle = np.arccos(-direction[1])  # Angle from downward

    return (
        float(rotation_axis[0] * angle),
        float(rotation_axis[1] * angle),
        float(rotation_axis[2] * angle),
    )


async def main():
    async with Nova() as nova, NovaRerunBridge(nova) as bridge:
        cell = nova.cell()

        # Initialize robots
        fanuc = await cell.ensure_virtual_robot_controller(
            "fanuc",
            models.VirtualControllerTypes.FANUC_MINUS_LR_MATE_200I_D7_L,
            models.Manufacturer.FANUC,
        )
        kuka = await cell.ensure_virtual_robot_controller(
            "kuka", models.VirtualControllerTypes.KUKA_MINUS_KR6_R700_2, models.Manufacturer.KUKA
        )
        abb = await cell.ensure_virtual_robot_controller(
            "abb", models.VirtualControllerTypes.ABB_MINUS_IRB1200_7, models.Manufacturer.ABB
        )
        yaskawa = await cell.ensure_virtual_robot_controller(
            "yaskawa", models.VirtualControllerTypes.YASKAWA_MINUS_GP7, models.Manufacturer.YASKAWA
        )

        # Set robot mountings
        for robot, pos in [
            (fanuc, ROBOT_POSITIONS["FANUC"]),
            (kuka, ROBOT_POSITIONS["KUKA"]),
            (abb, ROBOT_POSITIONS["ABB"]),
            (yaskawa, ROBOT_POSITIONS["YASKAWA"]),
        ]:
            await nova._api_client.virtual_robot_setup_api.set_virtual_robot_mounting(
                cell="cell",
                controller=robot.controller_id,
                id=pos.motion_group_id,
                coordinate_system=CoordinateSystem(
                    coordinate_system="world",
                    name="mounting",
                    reference_uid="",
                    position=pos.mounting,
                    rotation=pos.rotation,  # Use the robot-specific rotation
                ),
            )

        await asyncio.sleep(3)  # Wait for setup
        await bridge.setup_blueprint()

        robots = {"FANUC": fanuc, "KUKA": kuka, "ABB": abb, "YASKAWA": yaskawa}
        robot_sequence = ["FANUC", "KUKA", "YASKAWA", "ABB", "FANUC"]  # Complete circle
        await move_to_initial_positions(robots, bridge, ROBOT_POSITIONS)
        bridge.continue_after_sync()

        for i in range(len(robot_sequence) - 1):
            current_robot = robots[robot_sequence[i]]
            next_robot = robots[robot_sequence[i + 1]]
            current_pos = ROBOT_POSITIONS[robot_sequence[i]]
            next_pos = ROBOT_POSITIONS[robot_sequence[i + 1]]

            # Step 1: First robot picks up cube and moves to handover
            await asyncio.gather(
                pick_and_pass_cube(
                    current_robot,
                    bridge,
                    current_pos,
                    "pickup_and_handover",
                    current_pos.motion_group_id,
                    sync=True,
                ),
                # Step 2: Next robot moves to handover position
                pick_and_pass_cube(
                    next_robot,
                    bridge,
                    current_pos,  # Use current robot's handover position
                    "go_to_handover",
                    next_pos.motion_group_id,
                    sync=True,
                ),
            )

            bridge.continue_after_sync()

            # Step 3: Both robots move to home (after handover)
            await asyncio.gather(
                pick_and_pass_cube(
                    current_robot,
                    bridge,
                    current_pos,
                    "go_home_with_cube",
                    current_pos.motion_group_id,
                    sync=True,
                ),
                pick_and_pass_cube(
                    next_robot,
                    bridge,
                    next_pos,
                    "go_home_with_cube",
                    next_pos.motion_group_id,
                    sync=True,
                ),
            )

            # Step 4: Next robot places cube and returns home
            await pick_and_pass_cube(
                next_robot, bridge, next_pos, "place_cube", next_pos.motion_group_id
            )

            await asyncio.sleep(1)  # Small delay between transfers


if __name__ == "__main__":
    asyncio.run(main())



from scipy.spatial.transform import Rotation as R

from nova.api import models


def normalize_pose(pose: models.Pose2 | None = None) -> models.PlannerPose:
    """Convert pose to normalized format with Vector3d components."""
    # Default components
    default_position = models.Vector3d(x=0.0, y=0.0, z=0.0)
    default_orientation = models.Quaternion(x=0.0, y=0.0, z=0.0, w=0.0)

    # Use default pose if none provided
    if pose is None:
        return models.PlannerPose(position=default_position, orientation=default_orientation)

    # Handle position conversion
    if isinstance(pose.position, list):
        position = models.Vector3d(
            x=float(pose.position[0]), y=float(pose.position[1]), z=float(pose.position[2])
        )
    else:
        position = (
            pose.position
            if hasattr(pose, "position") and pose.position is not None
            else default_position
        )

    # Handle orientation conversion
    if isinstance(pose.orientation, list):
        orientation_rot = models.Vector3d(
            x=float(pose.orientation[0]), y=float(pose.orientation[1]), z=float(pose.orientation[2])
        )

        # Convert rotation vector to quaternion
        rot_vec = [orientation_rot.x, orientation_rot.y, orientation_rot.z]
        r = R.from_rotvec(rot_vec)
        quat = r.as_quat()  # [x, y, z, w]

        return models.PlannerPose(
            position=position,
            orientation=models.Quaternion(
                x=float(quat[0]), y=float(quat[1]), z=float(quat[2]), w=float(quat[3])
            ),
        )
    orientation = (
        pose.orientation
        if hasattr(pose, "orientation") and pose.orientation is not None
        else default_orientation
    )

    return models.PlannerPose(position=position, orientation=orientation)



import asyncio

import numpy as np
import rerun as rr
import trimesh
from wandelbots_api_client.models import (
    CoordinateSystem,
    RobotTcp,
    RotationAngles,
    RotationAngleTypes,
    Vector3d,
)

from nova import MotionSettings
from nova.actions import collision_free, linear
from nova.api import models
from nova.core.exceptions import PlanTrajectoryFailed
from nova.core.nova import Nova
from nova.types import Pose
from nova_rerun_bridge import NovaRerunBridge

"""
Simple example to demonstrate how to add a welding part to the collision world and move the robot to a two seams.
"""


async def load_and_transform_mesh(filepath: str, pose: models.Pose2) -> trimesh.Geometry:
    """Load mesh and transform to desired position."""
    scene = trimesh.load_mesh(filepath, file_type="stl")

    # Create transformation matrix from Pose2
    transform = np.eye(4)
    transform[:3, 3] = pose.position
    scene.apply_transform(transform)
    return scene


async def log_mesh_to_rerun(scene: trimesh.Trimesh) -> None:
    """Log mesh to rerun visualization."""
    vertices = scene.vertices
    faces = scene.faces
    vertex_normals = scene.vertex_normals
    vertex_colors = np.ones((len(vertices), 3), dtype=np.float32)

    rr.log(
        "motion/welding_benchmark",
        rr.Mesh3D(
            vertex_positions=vertices,
            triangle_indices=faces,
            vertex_normals=vertex_normals,
            albedo_factor=vertex_colors,
        ),
        static=True,
    )


async def add_mesh_to_collision_world(
    collision_api, cell_name: str, scene: trimesh.Trimesh, collider_name: str = "welding_part"
) -> models.Collider:
    """Add mesh as convex hull to collision world."""
    # Create convex hull
    convex_hull = scene.convex_hull

    # Create collider from convex hull vertices
    mesh_collider = models.Collider(
        shape=models.ColliderShape(
            models.ConvexHull2(vertices=convex_hull.vertices.tolist(), shape_type="convex_hull")
        ),
        margin=10,  # add 10mm margin to the convex hull
    )

    await collision_api.store_collider(
        cell=cell_name, collider=collider_name, collider2=mesh_collider
    )
    return mesh_collider


async def build_collision_world(
    nova: Nova, cell_name: str, robot_setup: models.OptimizerSetup, additional_colliders: dict = {}
) -> str:
    """Build collision world with robot, environment and optional additional colliders.

    Args:
        nova: Nova instance
        cell_name: Name of the cell
        robot_setup: Robot optimizer setup
        additional_colliders: Optional dictionary of additional colliders to add
    """
    collision_api = nova._api_client.store_collision_components_api
    scene_api = nova._api_client.store_collision_scenes_api

    # define robot base
    base_collider = models.Collider(
        shape=models.ColliderShape(models.Cylinder2(radius=200, height=300, shape_type="cylinder")),
        pose=models.Pose2(position=[0, 0, -155]),
    )
    await collision_api.store_collider(cell=cell_name, collider="base", collider2=base_collider)

    # define floor
    floor_collider = models.Collider(
        shape=models.ColliderShape(
            models.Box2(size_x=2000, size_y=2000, size_z=10, shape_type="box", box_type="FULL")
        ),
        pose=models.Pose2(position=[0, 0, -310]),
    )
    await collision_api.store_collider(cell=cell_name, collider="floor", collider2=floor_collider)

    # define TCP collider geometry
    tool_collider = models.Collider(
        shape=models.ColliderShape(
            models.Box2(size_x=5, size_y=5, size_z=100, shape_type="box", box_type="FULL")
        ),
        pose=models.Pose2(position=[0, 0, 50]),
    )
    await collision_api.store_collision_tool(
        cell=cell_name, tool="tool_box", request_body={"tool_collider": tool_collider}
    )

    # define robot link geometries
    robot_link_colliders = await collision_api.get_default_link_chain(
        cell=cell_name, motion_group_model=robot_setup.motion_group_type
    )
    await collision_api.store_collision_link_chain(
        cell=cell_name, link_chain="robot_links", collider=robot_link_colliders
    )

    # Prepare colliders dictionary
    colliders = {"base": base_collider, "floor": floor_collider}

    # Add additional colliders if provided
    if additional_colliders:
        colliders.update(additional_colliders)

    # assemble scene
    scene = models.CollisionScene(
        colliders=colliders,
        motion_groups={
            robot_setup.motion_group_type: models.CollisionMotionGroup(
                tool={"tool_geometry": tool_collider}, link_chain=robot_link_colliders
            )
        },
    )
    scene_id = "collision_scene"
    await scene_api.store_collision_scene(
        cell_name, scene_id, models.CollisionSceneAssembly(scene=scene)
    )
    return scene_id


async def calculate_seam_poses(mesh_pose: models.Pose2) -> tuple[Pose, Pose, Pose, Pose]:
    """Calculate seam poses relative to the mesh pose using @ operator.

    Args:
        mesh_pose: Position and orientation of the welding piece
    Returns:
        tuple containing start and end poses for both seams
    """
    # Convert mesh_pose to Pose for @ operator usage
    mesh_transform = Pose(mesh_pose)

    # Define seams in local coordinates (relative to mesh center)
    local_seam1_start = Pose((150, -6, 3, -np.pi / 2 - np.pi / 4, 0, 0))  # -135° around X
    local_seam1_end = Pose((30, -6, 3, -np.pi / 2 - np.pi / 4, 0, 0))
    local_seam2_start = Pose((150, 6, 3, np.pi / 2 + np.pi / 4, 0, 0))  # 135° around X
    local_seam2_end = Pose((30, 6, 3, np.pi / 2 + np.pi / 4, 0, 0))

    # Transform to global coordinates using @ operator
    seam1_start = mesh_transform @ local_seam1_start
    seam1_end = mesh_transform @ local_seam1_end
    seam2_start = mesh_transform @ local_seam2_start
    seam2_end = mesh_transform @ local_seam2_end

    return seam1_start, seam1_end, seam2_start, seam2_end


async def test():
    async with Nova() as nova, NovaRerunBridge(nova) as bridge:
        await bridge.setup_blueprint()

        # Define position for the welding part
        mesh_pose = models.Pose2(
            position=[500, 0, -300], orientation=[0, 0, 0]
        )  # in front of robot, on floor

        # Load and transform mesh
        scene = await load_and_transform_mesh(
            "nova_rerun_bridge/example_data/Welding_Benchmark_USA_01.stl", mesh_pose
        )

        # Log to rerun
        await log_mesh_to_rerun(scene)

        cell = nova.cell()
        controller = await cell.ensure_virtual_robot_controller(
            "ur10",
            models.VirtualControllerTypes.UNIVERSALROBOTS_MINUS_UR10E,
            models.Manufacturer.UNIVERSALROBOTS,
        )

        await nova._api_client.virtual_robot_setup_api.set_virtual_robot_mounting(
            cell="cell",
            controller=controller.controller_id,
            id=0,
            coordinate_system=CoordinateSystem(
                coordinate_system="world",
                name="mounting",
                reference_uid="",
                position=Vector3d(x=0, y=0, z=0),
                rotation=RotationAngles(
                    angles=[0, 0, 0], type=RotationAngleTypes.EULER_ANGLES_EXTRINSIC_XYZ
                ),
            ),
        )

        await nova._api_client.virtual_robot_setup_api.add_virtual_robot_tcp(
            cell="cell",
            controller="ur10",
            id=0,
            robot_tcp=RobotTcp(
                id="torch",
                position=Vector3d(x=0, y=0, z=100),
                rotation=RotationAngles(
                    angles=[0, 0, 0], type=RotationAngleTypes.EULER_ANGLES_EXTRINSIC_XYZ
                ),
            ),
        )

        await asyncio.sleep(5)

        # Connect to the controller and activate motion groups
        async with controller[0] as motion_group:
            await bridge.log_saftey_zones(motion_group)

            tcp = "torch"

            robot_setup: models.OptimizerSetup = await motion_group._get_optimizer_setup(tcp=tcp)

            # Add mesh to collision world
            mesh_collider = await add_mesh_to_collision_world(
                nova._api_client.store_collision_components_api, "cell", scene
            )

            # Build collision world with welding part included
            collision_scene_id = await build_collision_world(
                nova, "cell", robot_setup, additional_colliders={"welding_part": mesh_collider}
            )
            scene_api = nova._api_client.store_collision_scenes_api
            collision_scene = await scene_api.get_stored_collision_scene(
                cell="cell", scene=collision_scene_id
            )
            await bridge.log_collision_scene(collision_scene_id)

            # Calculate seam positions based on mesh pose
            seam1_start, seam1_end, seam2_start, seam2_end = await calculate_seam_poses(mesh_pose)

            # Define approach offset in local coordinates
            approach_offset = Pose((0, 0, -60, 0, 0, 0))

            # Create approach and departure poses using @ operator
            seam1_approach = seam1_start @ approach_offset
            seam1_departure = seam1_end @ approach_offset
            seam2_approach = seam2_start @ approach_offset
            seam2_departure = seam2_end @ approach_offset

            try:
                welding_actions = [
                    # First seam
                    collision_free(
                        target=seam1_approach,
                        collision_scene=collision_scene,
                        settings=MotionSettings(tcp_velocity_limit=30),
                    ),
                    linear(
                        target=seam1_start,
                        settings=MotionSettings(tcp_velocity_limit=30, blend_radius=10),
                    ),
                    linear(
                        target=seam1_end,
                        settings=MotionSettings(tcp_velocity_limit=30, blend_radius=10),
                    ),
                    linear(
                        target=seam1_departure,
                        settings=MotionSettings(tcp_velocity_limit=30, blend_radius=10),
                    ),
                    # Move to second seam
                    collision_free(
                        target=seam2_approach,
                        collision_scene=collision_scene,
                        settings=MotionSettings(tcp_velocity_limit=30),
                    ),
                    # Second seam with collision checking
                    linear(
                        target=seam2_start,
                        settings=MotionSettings(tcp_velocity_limit=30, blend_radius=10),
                    ),
                    linear(
                        target=seam2_end,
                        settings=MotionSettings(tcp_velocity_limit=30, blend_radius=10),
                    ),
                    linear(
                        target=seam2_departure,
                        settings=MotionSettings(tcp_velocity_limit=30, blend_radius=10),
                    ),
                    collision_free(
                        target=(0, -np.pi / 2, np.pi / 2, 0, 0, 0),
                        collision_scene=collision_scene,
                        settings=MotionSettings(tcp_velocity_limit=30),
                    ),
                ]

                trajectory_plan_combined = await motion_group.plan(
                    welding_actions,
                    tcp=tcp,
                    start_joint_position=(0, -np.pi / 2, np.pi / 2, 0, 0, 0),
                )

                await bridge.log_actions(welding_actions)
                await bridge.log_trajectory(trajectory_plan_combined, tcp, motion_group)

            except PlanTrajectoryFailed as e:
                await bridge.log_trajectory(e.error.joint_trajectory, tcp, motion_group)
                await bridge.log_error_feedback(e.error.error_feedback)
                raise


if __name__ == "__main__":
    asyncio.run(test())



import json
from pathlib import Path
from typing import Any

from nova.api import models


def _convert_sets_to_lists(obj: Any) -> Any:
    """Convert sets to lists in nested structures."""
    if isinstance(obj, set):
        return list(obj)
    if isinstance(obj, dict):
        return {k: _convert_sets_to_lists(v) for k, v in obj.items()}
    if isinstance(obj, list):
        return [_convert_sets_to_lists(item) for item in obj]
    return obj


def record_motion_data(
    motion_id: str,
    model_from_controller: str,
    motion_group: str,
    optimizer_config: models.OptimizerSetup,
    trajectory: list[models.TrajectorySample],
    collision_scenes: dict[str, models.CollisionScene],
    output_file: Path,
) -> None:
    """Record motion data to a JSON file for testing."""
    data = {
        "motion_id": motion_id,
        "model_from_controller": model_from_controller,
        "motion_group": motion_group,
        "optimizer_config": _convert_sets_to_lists(optimizer_config.model_dump()),
        "trajectory": [_convert_sets_to_lists(t.model_dump()) for t in trajectory],
        "collision_scenes": {
            k: _convert_sets_to_lists(v.model_dump()) for k, v in collision_scenes.items()
        },
    }

    output_file.parent.mkdir(parents=True, exist_ok=True)
    output_file.write_text(json.dumps(data, indent=2))



import asyncio
from typing import AsyncIterable, AsyncIterator, Callable, Generic, TypeVar

In = TypeVar("In")
Out = TypeVar("Out")


class StreamExtractor(Generic[In, Out]):
    def __init__(
        self,
        wrapped: Callable[[AsyncIterator[In]], AsyncIterable[Out]],
        stop_selector: Callable[[In], bool] | None = None,
    ):
        self._in_queue: asyncio.Queue[In | None] = asyncio.Queue()
        self._wrapped = wrapped
        self._stop_selector = stop_selector or (lambda x: x is None)

    def __call__(self, in_stream: AsyncIterable[In]) -> AsyncIterable[Out]:
        async def in_wrapper(in_stream_) -> AsyncIterator[In]:
            async for in_value in in_stream_:
                if self._stop_selector(in_value):
                    self._in_queue.put_nowait(None)
                else:
                    self._in_queue.put_nowait(in_value)
                yield in_value

        return self._wrapped(in_wrapper(in_stream))

    def __aiter__(self) -> AsyncIterator[In]:
        return self

    async def __anext__(self) -> In:
        value = await self._in_queue.get()
        if value is None:
            raise StopAsyncIteration
        return value



from __future__ import annotations

import asyncio
import functools
import time
from abc import ABC
from enum import Enum
from typing import AsyncGenerator, TypeVar

import wandelbots_api_client as wb
from decouple import config

from nova.auth.auth_config import Auth0Config
from nova.auth.authorization import Auth0DeviceAuthorization
from nova.core import logger
from nova.core.env_handler import set_key
from nova.core.exceptions import LoadPlanFailed, PlanTrajectoryFailed
from nova.core.robot_cell import ConfigurablePeriphery, Device
from nova.version import version as pkg_version

T = TypeVar("T")

INTERNAL_CLUSTER_NOVA_API = "http://api-gateway.wandelbots.svc.cluster.local:8080"


class ComparisonType(Enum):
    COMPARISON_TYPE_EQUAL = "COMPARISON_TYPE_EQUAL"
    COMPARISON_TYPE_GREATER = "COMPARISON_TYPE_GREATER"
    COMPARISON_TYPE_LESS = "COMPARISON_TYPE_LESS"


def intercept(api_instance: T, gateway: "ApiGateway") -> T:
    """Extend api interface classes to add logging and token validation"""

    class Interceptor:
        def __init__(self, instance: T):
            self._instance = instance

        def __getattr__(self, name):
            # Retrieve the original attribute
            original_attr = getattr(self._instance, name)

            # If it's not callable, return it as is
            if not callable(original_attr):
                return original_attr

            # Wrap async callables
            if asyncio.iscoroutinefunction(original_attr):

                @functools.wraps(original_attr)
                async def async_wrapper(*args, **kwargs):
                    await gateway._ensure_valid_token()
                    start = time.time()
                    try:
                        return await original_attr(*args, **kwargs)
                    except Exception as e:
                        logger.error(f"API CALL: {name} failed with error: {e}")
                        logger.debug(f"API CALL FAILED: {name} with args={args}, kwargs={kwargs}")
                        raise e
                    finally:
                        duration = time.time() - start
                        logger.info(f"API CALL: {name} took {duration:.2f} seconds")
                        logger.debug(f"API CALL: {name} with args={args}, kwargs={kwargs}")

                return async_wrapper

            # Wrap sync callables
            @functools.wraps(original_attr)
            def sync_wrapper(*args, **kwargs):
                start = time.time()
                try:
                    return original_attr(*args, **kwargs)
                except Exception as e:
                    logger.error(f"API CALL: {name} failed with error: {e}")
                    raise e
                finally:
                    duration = time.time() - start
                    logger.info(f"API CALL: {name} took {duration:.2f} seconds")
                    logger.debug(f"API CALL: {name} with args={args}, kwargs={kwargs}")

            return sync_wrapper

    # we ignore the type error here because
    # we want the return type to be the same as the original api instance to not break typing support
    return Interceptor(api_instance)  # type: ignore[return-value]


class ApiGateway:
    def __init__(
        self,
        *,
        host: str | None = None,
        username: str | None = None,
        password: str | None = None,
        access_token: str | None = None,
        version: str = "v1",
        verify_ssl: bool = True,
        auth0_config: Auth0Config | None = None,
    ):
        if host is None:
            host = config("NOVA_API", default=INTERNAL_CLUSTER_NOVA_API)

        if username is None:
            username = config("NOVA_USERNAME", default=None)

        if password is None:
            password = config("NOVA_PASSWORD", default=None)

        if access_token is None:
            access_token = config("NOVA_ACCESS_TOKEN", default=None)

        self._version = version
        self._verify_ssl = verify_ssl
        self._validating_token = False
        self._has_valid_token = False

        # Access token has more prio than username and password if both are provided at the same time, set username and
        # password to None
        if access_token is not None:
            username = None
            password = None

        self._auth0 = None
        auth0_config = auth0_config or Auth0Config.from_env()
        if auth0_config.is_complete():
            self._auth0 = Auth0DeviceAuthorization(auth0_config=auth0_config)

        self._host = self._host_with_prefix(host=host)
        stripped_host = self._host.rstrip("/")
        api_client_config = wb.Configuration(
            host=f"{stripped_host}/api/{version}",
            username=username,
            password=password,
            access_token=access_token,
        )
        api_client_config.verify_ssl = verify_ssl

        self._access_token = access_token
        self._username = username
        self._password = password

        self._init_api_client()

    def _init_api_client(self):
        """Initialize or reinitialize the API client with current credentials"""
        stripped_host = self._host.rstrip("/")
        api_client_config = wb.Configuration(
            host=f"{stripped_host}/api/{self._version}",
            username=self._username,
            password=self._password,
            access_token=self._access_token,
        )
        api_client_config.verify_ssl = self._verify_ssl

        self._api_client = wb.ApiClient(configuration=api_client_config)
        self._api_client.user_agent = f"Wandelbots-Nova-Python-SDK/{pkg_version}"

        # Use the intercept function to wrap each API client
        self.system_api = intercept(wb.SystemApi(api_client=self._api_client), self)
        self.controller_api = intercept(wb.ControllerApi(api_client=self._api_client), self)
        self.motion_group_api = intercept(wb.MotionGroupApi(api_client=self._api_client), self)
        self.motion_api = intercept(wb.MotionApi(api_client=self._api_client), self)
        self.motion_group_infos_api = intercept(
            wb.MotionGroupInfosApi(api_client=self._api_client), self
        )
        self.motion_group_kinematic_api = intercept(
            wb.MotionGroupKinematicApi(api_client=self._api_client), self
        )
        self.store_collision_components_api = intercept(
            wb.StoreCollisionComponentsApi(api_client=self._api_client), self
        )
        self.store_collision_scenes_api = intercept(
            wb.StoreCollisionScenesApi(api_client=self._api_client), self
        )
        self.virtual_robot_api = intercept(wb.VirtualRobotApi(api_client=self._api_client), self)
        self.virtual_robot_behavior_api = intercept(
            wb.VirtualRobotBehaviorApi(api_client=self._api_client), self
        )
        self.virtual_robot_mode_api = intercept(
            wb.VirtualRobotModeApi(api_client=self._api_client), self
        )
        self.virtual_robot_setup_api = intercept(
            wb.VirtualRobotSetupApi(api_client=self._api_client), self
        )
        self.controller_ios_api = intercept(wb.ControllerIOsApi(api_client=self._api_client), self)

        logger.debug(f"NOVA API client initialized with user agent {self._api_client.user_agent}")

    async def close(self):
        return await self._api_client.close()

    async def _ensure_valid_token(self):
        """Ensure we have a valid access token, requesting a new one if needed"""
        if not self._auth0 or self._validating_token or self._has_valid_token:
            return
        try:
            # Test token with a direct API call without interception
            self._validating_token = True
            async with wb.ApiClient(self._api_client.configuration) as client:
                api = wb.SystemApi(client)
                await api.get_system_version()
                self._has_valid_token = True
        except Exception as e:
            if "401" in str(e) or "403" in str(e):
                logger.info("Access token expired, starting device authorization flow")
                self._auth0.request_device_code()
                self._auth0.display_user_instructions()

                new_token = await self._auth0.poll_token_endpoint()
                self._access_token = new_token
                self._username = None
                self._password = None

                # Store the new token in .env file
                set_key("NOVA_ACCESS_TOKEN", new_token)

                # Update the existing API client configuration with the new token
                self._api_client.configuration.access_token = new_token
                self._api_client.configuration.username = None
                self._api_client.configuration.password = None

                # Reinitialize all API clients with the new configuration
                self._init_api_client()

                logger.info("Successfully updated access token and reinitialized API clients")
        finally:
            self._validating_token = False

    @staticmethod
    def _host_with_prefix(host: str) -> str:
        """
        The protocol prefix is required for the API client to work properly.
        This method adds the 'http://' prefix if it is missing.
        For all wandelbots.io virtual instances the prefix will 'https://'.
        """
        is_wabo_host = "wandelbots.io" in host
        if host.startswith("http") and not is_wabo_host:
            return host
        if host.startswith("http") and is_wabo_host:
            return host.replace("http://", "https://")
        if is_wabo_host:
            return f"https://{host}"
        return f"http://{host}"

    @property
    def host(self) -> str:
        return self._host

    @property
    def access_token(self) -> str | None:
        return self._access_token

    @property
    def username(self) -> str | None:
        return self._username

    @property
    def password(self) -> str | None:
        return self._password

    async def stream_robot_controller_state(
        self, cell: str, controller_id: str, response_rate: int = 200
    ) -> AsyncGenerator[wb.models.RobotControllerState, None]:
        """
        Stream the robot controller state.
        """
        async for state in self.controller_api.stream_robot_controller_state(
            cell=cell, controller=controller_id, response_rate=response_rate
        ):
            yield state

    async def activate_motion_group(self, cell: str, motion_group_id: str):
        await self.motion_group_api.activate_motion_group(cell=cell, motion_group=motion_group_id)

    async def activate_all_motion_groups(self, cell: str, controller: str) -> list[str]:
        """
        Activate all motion groups for the given cell and controller.
        Returns the id of the activated motion groups.
        """
        activate_all_motion_groups_response = (
            await self.motion_group_api.activate_all_motion_groups(cell=cell, controller=controller)
        )
        motion_groups = activate_all_motion_groups_response.instances
        return [mg.motion_group for mg in motion_groups]

    async def list_controller_io_descriptions(
        self, cell: str, controller: str, ios: list[str]
    ) -> list[wb.models.IODescription]:
        if not ios:
            ios = []

        response = await self.controller_ios_api.list_io_descriptions(
            cell=cell, controller=controller, ios=ios
        )
        return response.io_descriptions

    async def read_controller_io(self, cell: str, controller: str, io: str) -> float | bool | int:
        response = await self.controller_ios_api.list_io_values(
            cell=cell, controller=controller, ios=[io]
        )

        found_io = response.io_values[0]
        if found_io.boolean_value is not None:
            return bool(found_io.boolean_value)
        if found_io.integer_value is not None:
            return int(found_io.integer_value)
        if found_io.floating_value is not None:
            return float(found_io.floating_value)
        raise ValueError(
            f"IO value for {io} is of an unexpected type. Expected bool, int or float."
        )

    async def write_controller_io(
        self, cell: str, controller: str, io: str, value: bool | int | float
    ):
        if isinstance(value, bool):
            io_value = wb.models.IOValue(io=io, boolean_value=value)
        elif isinstance(value, int):
            io_value = wb.models.IOValue(io=io, integer_value=str(value))
        elif isinstance(value, float):
            io_value = wb.models.IOValue(io=io, floating_value=value)
        else:
            raise ValueError(f"Invalid value type {type(value)}. Expected bool, int or float.")

        await self.controller_ios_api.set_output_values(
            cell=cell, controller=controller, io_value=[io_value]
        )

    async def wait_for_bool_io(self, cell: str, controller: str, io: str, value: bool):
        await self.controller_ios_api.wait_for_io_event(
            cell=cell,
            controller=controller,
            io=io,
            comparison_type=ComparisonType.COMPARISON_TYPE_EQUAL,
            boolean_value=value,
        )

    async def list_controllers(self, *, cell: str) -> list[wb.models.ControllerInstance]:
        response = await self.controller_api.list_controllers(cell=cell)
        return response.instances

    async def get_controller_instance(
        self, *, cell: str, name: str
    ) -> wb.models.ControllerInstance | None:
        controllers = await self.list_controllers(cell=cell)
        return next((c for c in controllers if c.controller == name), None)

    async def get_current_robot_controller_state(
        self, *, cell: str, controller_id: str
    ) -> wb.models.RobotControllerState:
        return await self.controller_api.get_current_robot_controller_state(
            cell=cell, controller=controller_id
        )

    async def add_robot_controller(
        self, cell: str, robot_controller: wb.models.RobotController, timeout: int = 25
    ):
        """
        Add a robot controller to the specified cell.
        Args:
            cell: The cell to add the controller to.
            robot_controller: The robot controller to add.
            timeout: The timeout in seconds for the operation.
        """
        await self.controller_api.add_robot_controller(
            cell=cell, robot_controller=robot_controller, completion_timeout=timeout
        )

    async def delete_robot_controller(
        self, *, cell: str, controller: str, completion_timeout: int = 25
    ) -> None:
        await self.controller_api.delete_robot_controller(
            cell=cell, controller=controller, completion_timeout=completion_timeout
        )

    async def wait_for_controller_ready(self, cell: str, name: str, timeout: int = 25) -> None:
        """
        Wait until the given controller has finished initializing or until timeout.
        Args:
            cell: The cell to check.
            name: The name of the controller.
            timeout: The timeout in seconds.
        """
        iteration = 0
        while iteration < timeout:
            controller = await self.get_controller_instance(cell=cell, name=name)
            if controller is None:
                logger.info(f"Controller not found: {cell}/{name}")
            elif controller.has_error:
                logger.error(f"Controller {cell}/{name} has error: {controller.error_details}")
            else:
                logger.info(f"Controller {cell}/{name} is ready")
                return

            await asyncio.sleep(1)
            iteration += 1

        raise TimeoutError(f"Timeout waiting for {cell}/{name} controller availability")

    async def add_virtual_robot_controller(
        self,
        cell: str,
        name: str,
        controller_type: wb.models.VirtualControllerTypes,
        controller_manufacturer: wb.models.Manufacturer,
        timeout: int = 25,
        position: str | None = None,
    ) -> wb.models.ControllerInstance:
        """
        Add a virtual robot controller to the cell and wait until it's ready.
        Returns the resulting ControllerInstance.
        Args:
            cell: The cell to add the controller to.
            name: The name of the controller.
            controller_type: The type of the controller.
            controller_manufacturer: The manufacturer of the controller.
            timeout: The timeout for waiting for the controller to be ready.
            position: The position of the controller in the cell.

        Returns:
            The resulting ControllerInstance.
        """
        if position is None:
            position = "[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]"  # fallback

        robot_controller = wb.models.RobotController(
            configuration=wb.models.RobotControllerConfiguration(
                wb.models.VirtualController(
                    manufacturer=controller_manufacturer, type=controller_type, position=position
                )
            ),
            name=name,
        )

        # Step 1: Add the controller
        await self.add_robot_controller(
            cell=cell, robot_controller=robot_controller, timeout=timeout
        )
        # Step 2: Wait for it to become ready
        await self.wait_for_controller_ready(cell=cell, name=name, timeout=timeout)

        # Step 3: Retrieve the instance
        controller_instance = await self.get_controller_instance(cell=cell, name=name)
        if controller_instance is None:
            raise ValueError(f"Controller not found after creation: {name}")

        return controller_instance

    async def plan_trajectory(
        self, cell: str, motion_group_id: str, request: wb.models.PlanTrajectoryRequest
    ) -> wb.models.JointTrajectory:
        """
        Plan a trajectory for the given motion group.
        """

        plan_trajectory_response = await self.motion_api.plan_trajectory(
            cell=cell, plan_trajectory_request=request
        )
        if isinstance(
            plan_trajectory_response.response.actual_instance,
            wb.models.PlanTrajectoryFailedResponse,
        ):
            # TODO: handle partially executable path
            raise PlanTrajectoryFailed(
                plan_trajectory_response.response.actual_instance, motion_group_id
            )
        return plan_trajectory_response.response.actual_instance

    async def load_planned_motion(
        self, cell: str, motion_group_id: str, joint_trajectory: wb.models.JointTrajectory, tcp: str
    ) -> wb.models.PlanSuccessfulResponse:
        load_plan_response = await self.motion_api.load_planned_motion(
            cell=cell,
            planned_motion=wb.models.PlannedMotion(
                motion_group=motion_group_id,
                times=joint_trajectory.times,
                joint_positions=joint_trajectory.joint_positions,
                locations=joint_trajectory.locations,
                tcp=tcp,
            ),
        )

        if (
            load_plan_response.plan_failed_on_trajectory_response is not None
            or load_plan_response.plan_failed_on_trajectory_response is not None
        ):
            raise LoadPlanFailed(load_plan_response)

        return load_plan_response.plan_successful_response

    def stream_move_to_trajectory_via_join_ptp(
        self,
        cell: str,
        motion_id: str,
        location_on_trajectory: int,
        joint_velocity_limits: wb.models.Joints | None = None,
    ) -> AsyncGenerator[wb.models.StreamMoveResponse, None]:
        return self.motion_api.stream_move_to_trajectory_via_joint_ptp(
            cell=cell,
            motion=motion_id,
            location_on_trajectory=location_on_trajectory,
            # limit_override_joint_velocity_limits_joints=joint_velocity_limits,
        )

    async def stop_motion(self, cell: str, motion_id: str):
        await self.motion_api.stop_execution(cell=cell, motion=motion_id)

    async def get_motion_group_state(
        self, cell: str, motion_group_id: str, tcp: str | None = None
    ) -> wb.models.MotionGroupStateResponse:
        return await self.motion_group_infos_api.get_current_motion_group_state(
            cell=cell, motion_group=motion_group_id, tcp=tcp
        )

    async def list_tcps(self, cell: str, motion_group_id: str) -> wb.models.ListTcpsResponse:
        return await self.motion_group_infos_api.list_tcps(cell=cell, motion_group=motion_group_id)

    async def get_active_tcp(self, cell: str, motion_group_id: str) -> wb.models.RobotTcp:
        return await self.motion_group_infos_api.get_active_tcp(
            cell=cell, motion_group=motion_group_id
        )

    async def get_optimizer_config(
        self, cell: str, motion_group_id: str, tcp: str
    ) -> wb.models.OptimizerSetup:
        return await self.motion_group_infos_api.get_optimizer_configuration(
            cell=cell, motion_group=motion_group_id, tcp=tcp
        )

    async def get_joint_number(self, cell: str, motion_group_id: str) -> int:
        spec = await self.motion_group_infos_api.get_motion_group_specification(
            cell=cell, motion_group=motion_group_id
        )
        return len(spec.mechanical_joint_limits)

    async def plan_collision_free_ptp(
        self, cell: str, motion_group_id: str, request: wb.models.PlanCollisionFreePTPRequest
    ):
        plan_result = await self.motion_api.plan_collision_free_ptp(
            cell=cell, plan_collision_free_ptp_request=request
        )

        if isinstance(plan_result.response.actual_instance, wb.models.PlanTrajectoryFailedResponse):
            raise PlanTrajectoryFailed(plan_result.response.actual_instance, motion_group_id)
        return plan_result.response.actual_instance


class NovaDevice(ConfigurablePeriphery, Device, ABC, is_abstract=True):
    class Configuration(ConfigurablePeriphery.Configuration):
        nova_api: str
        nova_access_token: str | None = None
        nova_username: str | None = None
        nova_password: str | None = None

    _nova_api: ApiGateway

    def __init__(self, configuration: Configuration, **kwargs):
        self._nova_api = ApiGateway(
            host=configuration.nova_api,
            access_token=configuration.nova_access_token,
            username=configuration.nova_username,
            password=configuration.nova_password,
        )
        super().__init__(configuration, **kwargs)



import asyncio
from typing import AsyncIterable, cast

import wandelbots_api_client as wb

from nova.actions import Action, CombinedActions, MovementController, MovementControllerContext
from nova.actions.motions import CollisionFreeMotion, Motion
from nova.api import models
from nova.core import logger
from nova.core.exceptions import InconsistentCollisionScenes
from nova.core.gateway import ApiGateway
from nova.core.movement_controller import move_forward
from nova.core.robot_cell import AbstractRobot
from nova.types import InitialMovementStream, LoadPlanResponse, MovementResponse, Pose, RobotState
from nova.utils import StreamExtractor

MAX_JOINT_VELOCITY_PREPARE_MOVE = 0.2
START_LOCATION_OF_MOTION = 0.0


def compare_collision_scenes(scene1: wb.models.CollisionScene, scene2: wb.models.CollisionScene):
    if scene1.colliders != scene2.colliders:
        return False

    # Compare motion groups
    if scene1.motion_groups != scene2.motion_groups:
        return False

    return True


# TODO: when collision scene is different in different motions
#  , we should plan them separately
def split_actions_into_batches(actions: list[Action]) -> list[list[Action]]:
    """
    Splits the list of actions into batches of actions and collision free motions.
    Actions are sent to plan_trajectory API and collision free motions are sent to plan_collision_free_ptp API.
    """
    batches: list[list[Action]] = []
    for action in actions:
        if (
            # Start a new batch if:
            not batches  # first action no batches yet
            or isinstance(action, CollisionFreeMotion)
            or isinstance(batches[-1][-1], CollisionFreeMotion)
        ):
            batches.append([action])
        else:
            batches[-1].append(action)
    return batches


def combine_trajectories(
    trajectories: list[wb.models.JointTrajectory],
) -> wb.models.JointTrajectory:
    """
    Combines multiple trajectories into one trajectory.
    """
    final_trajectory = trajectories[0]
    current_end_time = final_trajectory.times[-1]
    current_end_location = final_trajectory.locations[-1]

    for trajectory in trajectories[1:]:
        # Shift times and locations to continue from last endpoint
        shifted_times = [t + current_end_time for t in trajectory.times[1:]]  # Skip first point
        shifted_locations = [
            location + current_end_location for location in trajectory.locations[1:]
        ]  # Skip first point

        final_trajectory.times.extend(shifted_times)
        final_trajectory.joint_positions.extend(trajectory.joint_positions[1:])
        final_trajectory.locations.extend(shifted_locations)

        current_end_time = final_trajectory.times[-1]
        current_end_location = final_trajectory.locations[-1]

    return final_trajectory


class MotionGroup(AbstractRobot):
    """Manages motion planning and execution within a specified motion group."""

    def __init__(self, api_gateway: ApiGateway, cell: str, motion_group_id: str):
        """
        Initializes a new MotionGroup instance.

        Args:
            api_gateway (ApiGateway): The API gateway through which motion commands are sent.
            cell (str): The name or identifier of the robotic cell.
            motion_group_id (str): The identifier of the motion group.
        """
        self._api_gateway = api_gateway
        self._cell = cell
        self._motion_group_id = motion_group_id
        self._current_motion: str | None = None
        self._optimizer_setup: wb.models.OptimizerSetup | None = None
        super().__init__(id=motion_group_id)

    async def open(self):
        await self._api_gateway.activate_motion_group(
            cell=self._cell, motion_group_id=self._motion_group_id
        )
        return self

    async def close(self):
        # RPS-1174: when a motion group is deactivated, RAE closes all open connections
        #           this behaviour is not desired in some cases,
        #           so for now we will not deactivate for the user
        pass

    @property
    def motion_group_id(self) -> str:
        """
        Returns:
            str: The unique identifier for this motion group.
        """
        return self._motion_group_id

    @property
    def current_motion(self) -> str | None:
        # if not self._current_motion:
        #    raise ValueError("No MotionId attached. There is no planned motion available.")
        return self._current_motion

    async def _plan_with_collision_check(
        self,
        actions: list[Action],
        tcp: str,
        start_joint_position: tuple[float, ...] | None = None,
        optimizer_setup: wb.models.OptimizerSetup | None = None,
    ) -> wb.models.JointTrajectory:
        """
        This method plans a trajectory and checks for collisions.
        The collision check only happens if the actions have collision scene data.

        You must provide the exact same collision data into all the actions.
        Because the underlying API supports collision checks for the whole trajectory only.

        Raises:
            InconsistentCollisionScenes: If the collision scene data is not consistent across all actions

            Your actions should follow below rules to be considered consistent:
            1- They all should have the same collision scene data
            2- They all should have no collision data

            PlanTrajectoryFailed: If the trajectory planning failed including the collision check

        For more information about this API, please refer to the plan_trajectory in the API documentation.

        Args:
            actions: list of actions to plan, current supported actions are Motion and WriteActions
                     WriteAction you specify on your path is handled in a performant way.
                     Please check execute_trajectory.motion_command.set_io for more information.
            tcp:     The tool to use
            start_joint_position: The starting joint position, if none provided, current position of the robot is used
            optimizer_setup: The optimizer setup

        Returns: planned joint trajectory

        """
        # PREPARE THE REQUEST
        collision_scenes = self._validate_collision_scenes(actions)
        start_joint_position = start_joint_position or await self.joints()
        robot_setup = optimizer_setup or await self._get_optimizer_setup(tcp=tcp)

        motion_commands = CombinedActions(items=tuple(actions)).to_motion_command()  # type: ignore

        static_colliders = None
        collision_motion_group = None
        if collision_scenes and len(collision_scenes) > 0:
            static_colliders = collision_scenes[0].colliders

            motion_group_type = robot_setup.motion_group_type
            if (
                collision_scenes[0].motion_groups
                and motion_group_type in collision_scenes[0].motion_groups
            ):
                collision_motion_group = collision_scenes[0].motion_groups[motion_group_type]

        request = wb.models.PlanTrajectoryRequest(
            robot_setup=robot_setup,
            start_joint_position=list(start_joint_position),
            motion_commands=motion_commands,
            static_colliders=static_colliders,
            collision_motion_group=collision_motion_group,
        )

        return await self._api_gateway.plan_trajectory(
            cell=self._cell, motion_group_id=self.motion_group_id, request=request
        )

    def _validate_collision_scenes(self, actions: list[Action]) -> list[models.CollisionScene]:
        motion_count = len([action for action in actions if isinstance(action, Motion)])
        collision_scenes = [
            action.collision_scene
            for action in actions
            if isinstance(action, CollisionFreeMotion) and action.collision_scene is not None
        ]

        if len(collision_scenes) != 0 and len(collision_scenes) != motion_count:
            raise InconsistentCollisionScenes(
                "Only some of the actions have collision scene. Either specify it for all or none."
            )

        # If a collision scene is provided, the same should be provided for all the collision scene
        if len(collision_scenes) > 1:
            first_scene = collision_scenes[0]
            if not all(
                compare_collision_scenes(first_scene, scene) for scene in collision_scenes[1:]
            ):
                raise InconsistentCollisionScenes(
                    "All actions must use the same collision scene but some are different"
                )

        return collision_scenes

    # TODO: we get the optimizer setup from as an input because
    #  it has a velocity setting which is used in collision free movement, I need to double check this
    async def _plan_collision_free(
        self,
        action: CollisionFreeMotion,
        tcp: str,
        start_joint_position: list[float],
        optimizer_setup: wb.models.OptimizerSetup | None = None,
    ) -> wb.models.JointTrajectory:
        """
        This method plans a trajectory and avoids collisions.
        This means if there is a collision along the way to the target pose or joint positions,
        It will adjust the trajectory to avoid the collision.

        The collision check only happens if the action have collision scene data.

        For more information about this API, please refer to the plan_collision_free_ptp in the API documentation.

        Args:
            action: The target pose or joint positions to reach
            tcp:     The tool to use
            start_joint_position: The starting joint position, if none provided, current position of the robot is used
            optimizer_setup: The optimizer setup

        Returns: planned joint trajectory


        """
        target = wb.models.PlanCollisionFreePTPRequestTarget(
            **action.model_dump(exclude_unset=True)
        )
        robot_setup = optimizer_setup or await self._get_optimizer_setup(tcp=tcp)

        static_colliders = None
        collision_motion_group = None
        collision_scene = action.collision_scene
        if collision_scene and collision_scene.colliders:
            static_colliders = collision_scene.colliders

            if (
                collision_scene.motion_groups
                and robot_setup.motion_group_type in collision_scene.motion_groups
            ):
                collision_motion_group = collision_scene.motion_groups[
                    robot_setup.motion_group_type
                ]

        request: wb.models.PlanCollisionFreePTPRequest = wb.models.PlanCollisionFreePTPRequest(
            robot_setup=robot_setup,
            start_joint_position=start_joint_position,
            target=target,
            static_colliders=static_colliders,
            collision_motion_group=collision_motion_group,
        )

        return await self._api_gateway.plan_collision_free_ptp(
            cell=self._cell, motion_group_id=self.motion_group_id, request=request
        )

    async def _plan(
        self,
        actions: list[Action],
        tcp: str,
        start_joint_position: tuple[float, ...] | None = None,
        optimizer_setup: wb.models.OptimizerSetup | None = None,
    ) -> wb.models.JointTrajectory:
        if not actions:
            raise ValueError("No actions provided")

        current_joints = start_joint_position or await self.joints()
        robot_setup = optimizer_setup or await self._get_optimizer_setup(tcp=tcp)

        all_trajectories = []
        for batch in split_actions_into_batches(actions):
            if len(batch) == 0:
                raise ValueError("Empty batch of actions")

            if isinstance(batch[0], CollisionFreeMotion):
                motion: CollisionFreeMotion = cast(CollisionFreeMotion, batch[0])
                trajectory = await self._plan_collision_free(
                    action=motion,
                    tcp=tcp,
                    start_joint_position=list(current_joints),
                    optimizer_setup=robot_setup,
                )
                all_trajectories.append(trajectory)
                # the last joint position of this trajectory is the starting point for the next one
                current_joints = tuple(trajectory.joint_positions[-1].joints)
            else:
                trajectory = await self._plan_with_collision_check(
                    actions=batch,
                    tcp=tcp,
                    start_joint_position=current_joints,
                    optimizer_setup=robot_setup,
                )
                all_trajectories.append(trajectory)
                # the last joint position of this trajectory is the starting point for the next one
                current_joints = tuple(trajectory.joint_positions[-1].joints)

        return combine_trajectories(all_trajectories)

    # TODO: refactor and simplify code, tests are already there
    # TODO: split into batches when the collision scene changes in a batch of collision free motions

    async def _execute(
        self,
        joint_trajectory: wb.models.JointTrajectory,
        tcp: str,
        actions: list[Action],
        movement_controller: MovementController | None,
    ) -> AsyncIterable[MovementResponse]:
        if movement_controller is None:
            movement_controller = move_forward

        # Load planned trajectory
        load_plan_response = await self._load_planned_motion(joint_trajectory, tcp)

        # Move to start position
        number_of_joints = await self._api_gateway.get_joint_number(
            cell=self._cell, motion_group_id=self.motion_group_id
        )
        joints_velocities = [MAX_JOINT_VELOCITY_PREPARE_MOVE] * number_of_joints
        movement_stream = await self.move_to_start_position(joints_velocities, load_plan_response)

        # If there's an initial consumer, feed it the data
        async for move_to_response in movement_stream:
            # TODO: refactor
            if (
                move_to_response.state is None
                or move_to_response.state.motion_groups is None
                or len(move_to_response.state.motion_groups) == 0
                or move_to_response.move_response is None
                or move_to_response.move_response.current_location_on_trajectory is None
            ):
                continue

            yield move_to_response

        controller = movement_controller(
            MovementControllerContext(
                combined_actions=CombinedActions(items=tuple(actions)),  # type: ignore
                motion_id=load_plan_response.motion,
            )
        )

        def stop_condition(response: wb.models.ExecuteTrajectoryResponse) -> bool:
            instance = response.actual_instance
            # Stop when standstill indicates motion ended
            return (
                isinstance(instance, wb.models.Standstill)
                and instance.standstill.reason == wb.models.StandstillReason.REASON_MOTION_ENDED
            )

        execute_response_streaming_controller = StreamExtractor(controller, stop_condition)
        execution_task = asyncio.create_task(
            self._api_gateway.motion_api.execute_trajectory(
                cell=self._cell, client_request_generator=execute_response_streaming_controller
            )
        )
        async for execute_response in execute_response_streaming_controller:
            yield execute_response
        await execution_task

    async def _get_optimizer_setup(self, tcp: str) -> wb.models.OptimizerSetup:
        # TODO: mypy failed on main branch, need to check
        if self._optimizer_setup is None or self._optimizer_setup.tcp != tcp:  # type: ignore
            self._optimizer_setup = await self._api_gateway.get_optimizer_config(
                cell=self._cell, motion_group_id=self.motion_group_id, tcp=tcp
            )

        return self._optimizer_setup

    async def _load_planned_motion(
        self, joint_trajectory: wb.models.JointTrajectory, tcp: str
    ) -> wb.models.PlanSuccessfulResponse:
        return await self._api_gateway.load_planned_motion(
            cell=self._cell,
            motion_group_id=self.motion_group_id,
            joint_trajectory=joint_trajectory,
            tcp=tcp,
        )

    async def move_to_start_position(
        self, joint_velocities, load_plan_response: LoadPlanResponse
    ) -> InitialMovementStream:
        limit_override = wb.models.LimitsOverride()
        if joint_velocities is not None:
            limit_override.joint_velocity_limits = wb.models.Joints(joints=joint_velocities)

        return self._api_gateway.stream_move_to_trajectory_via_join_ptp(
            cell=self._cell,
            motion_id=load_plan_response.motion,
            location_on_trajectory=0,
            joint_velocity_limits=limit_override.joint_velocity_limits,
        )

    async def stop(self):
        logger.debug(f"Stopping motion of {self}...")
        try:
            if self._current_motion is None:
                raise ValueError("No motion to stop")
            await self._api_gateway.stop_motion(cell=self._cell, motion_id=self._current_motion)
            logger.debug(f"Motion {self.current_motion} stopped.")
        except ValueError as e:
            logger.debug(f"No motion to stop for {self}: {e}")

    async def get_state(self, tcp: str | None = None) -> RobotState:
        response = await self._api_gateway.get_motion_group_state(
            cell=self._cell, motion_group_id=self.motion_group_id, tcp=tcp
        )
        return RobotState(
            pose=Pose(response.state.tcp_pose), joints=tuple(response.state.joint_position.joints)
        )

    async def joints(self) -> tuple:
        state = await self.get_state()
        if state.joints is None:
            raise ValueError(
                f"No joint positions available for motion group {self._motion_group_id}"
            )
        return state.joints

    async def tcp_pose(self, tcp: str | None = None) -> Pose:
        state = await self.get_state(tcp=tcp)
        return state.pose

    async def tcps(self) -> list[wb.models.RobotTcp]:
        response = await self._api_gateway.list_tcps(
            cell=self._cell, motion_group_id=self.motion_group_id
        )
        return response.tcps if response.tcps else []

    async def tcp_names(self) -> list[str]:
        return [tcp.id for tcp in await self.tcps()]

    async def active_tcp(self) -> wb.models.RobotTcp:
        active_tcp = await self._api_gateway.get_active_tcp(
            cell=self._cell, motion_group_id=self.motion_group_id
        )
        return active_tcp

    async def active_tcp_name(self) -> str:
        active_tcp = await self.active_tcp()
        return active_tcp.id



from nova.cell.cell import Cell
from nova.cell.controllers import (
    abb_controller,
    fanuc_controller,
    kuka_controller,
    universal_robots_controller,
    virtual_controller,
    yaskawa_controller,
)

__all__ = [
    "Cell",
    "yaskawa_controller",
    "fanuc_controller",
    "universal_robots_controller",
    "kuka_controller",
    "abb_controller",
    "virtual_controller",
]



from dataclasses import dataclass, field
from pathlib import Path
from typing import Any

import numpy as np
import rerun as rr
from scipy.spatial.transform import Rotation

from nova.core.motion_group import MotionGroup
from nova_rerun_bridge import NovaRerunBridge
from nova_rerun_bridge.benchmark.robometrics_helper import convert_position


def _default_colors() -> dict[str, tuple[float, float, float, float]]:
    """Default color configuration for visualization."""
    return {
        "x_axis": (1.0, 0.125, 0.376, 1.0),  # #ff2060 - Red/Pink
        "y_axis": (0.125, 0.875, 0.502, 1.0),  # #20df80 - Green
        "z_axis": (0.125, 0.502, 1.0, 1.0),  # #2080ff - Blue
        "target": (0, 255, 0, 255),  # Green
    }


@dataclass
class VisualizationConfig:
    """Configuration for trajectory visualization."""

    arrow_length: float = 50.0  # mm
    point_radius: float = 5.0
    arrow_radius: float = 2.5
    colors: dict[str, tuple[float, float, float, float]] = field(default_factory=_default_colors)


async def log_successful_planning(
    bridge: NovaRerunBridge,
    strategy: str,
    collision_scene_id: str,
    problem: dict[str, Any],
    key: str,
    iteration: int,
    trajectory: Any,
    tcp: str,
    motion_group: MotionGroup,
    vis_config: VisualizationConfig = VisualizationConfig(),
) -> str:
    """Log successful planning attempt with visualizations.

    Args:
        bridge: NovaRerunBridge instance
        collision_scene_id: ID of the collision scene
        problem: Benchmark problem definition
        key: Scene identifier
        iteration: Current iteration number
        trajectory: Planned trajectory
        tcp: Tool Center Point identifier
        motion_group: Motion group instance
        vis_config: Visualization configuration
    """
    result_dir = Path("benchmark_results") / strategy / key
    result_dir.mkdir(parents=True, exist_ok=True)

    # Initialize rerun visualization
    recording_id = f"{strategy}_{key}_{iteration}"

    rr.init(application_id="nova", recording_id=recording_id, spawn=False)
    rr.save(str(result_dir / f"{recording_id}.rrd"))

    await bridge.setup_blueprint()
    await bridge.log_collision_scene(collision_scene_id)

    # Process goal pose
    goal_position = convert_position(problem["goal_pose"]["position_xyz"])
    goal_quat = problem["goal_pose"]["quaternion_wxyz"]

    # Log target position
    _log_target_position(goal_position, vis_config)

    # Log orientation arrows
    _log_orientation_arrows(goal_position, goal_quat, vis_config)

    # Log trajectory
    await bridge.log_trajectory(trajectory, tcp, motion_group)

    return f"{result_dir / f'{recording_id}.rrd'}"


def _log_target_position(position: list[float], config: VisualizationConfig) -> None:
    """Log target position as a 3D point."""
    rr.log(
        "motion/target_orientation",
        rr.Points3D(
            positions=[position], radii=[config.point_radius], colors=[config.colors["target"]]
        ),
    )


def _log_orientation_arrows(
    position: list[float], quaternion: list[float], config: VisualizationConfig
) -> None:
    """Log orientation as colored coordinate arrows."""
    w, x, y, z = quaternion
    rot = Rotation.from_quat([x, y, z, w])

    # Create arrow vectors for each axis
    arrow_vectors = [
        rot.apply([config.arrow_length, 0, 0]),  # X axis
        rot.apply([0, config.arrow_length, 0]),  # Y axis
        rot.apply([0, 0, config.arrow_length]),  # Z axis
    ]

    # Convert colors to numpy array
    coordinate_colors = np.array(
        [config.colors["x_axis"], config.colors["y_axis"], config.colors["z_axis"]]
    )

    rr.log(
        "motion/target_orientation",
        rr.Arrows3D(
            origins=[position] * 3,
            vectors=arrow_vectors,
            colors=coordinate_colors,
            radii=[config.arrow_radius] * 3,
        ),
        static=True,
    )



from __future__ import annotations

import asyncio
from enum import Enum

from nova.api import models
from nova.core.gateway import ApiGateway
from nova.core.robot_cell import Device, ValueType


class IOType(Enum):
    IO_TYPE_INPUT = "IO_TYPE_INPUT"
    IO_TYPE_OUTPUT = "IO_TYPE_OUTPUT"


class IOValueType(Enum):
    IO_VALUE_ANALOG_INTEGER = "IO_VALUE_ANALOG_INTEGER"
    IO_VALUE_ANALOG_FLOATING = "IO_VALUE_ANALOG_FLOATING"
    IO_VALUE_DIGITAL = "IO_VALUE_DIGITAL"


class IOAccess(Device):
    """Provides access to input and outputs via a dictionary-like style

    TODO:
        - Add listener to value changes
        - Handle integer masks
        - Read and check types based on the description
    """

    io_descriptions_cache: dict[str, dict[str, models.IODescription]] = {}

    def __init__(self, api_gateway: ApiGateway, cell: str, controller_id: str):
        super().__init__()
        self._api_gateway = api_gateway
        self._cell = cell
        self._controller_id = controller_id
        self._io_operation_in_progress = asyncio.Lock()

    async def get_io_descriptions(self) -> dict[str, models.IODescription]:
        cache = self.__class__.io_descriptions_cache
        if self._controller_id not in cache:
            # empty list fetches all
            io_descriptions = await self._api_gateway.list_controller_io_descriptions(
                cell=self._cell, controller=self._controller_id, ios=[]
            )
            cache[self._controller_id] = {
                description.id: description for description in io_descriptions
            }
        return cache[self._controller_id]

    @staticmethod
    def filter_io_descriptions(
        io_descriptions: dict[str, models.IODescription],
        filter_value_type: IOValueType | None = None,
        filter_type: IOType | None = None,
    ) -> list[str]:
        return [
            io.id
            for io in io_descriptions.values()
            if filter_value_type is None
            or (IOValueType(io.value_type) == filter_value_type and IOType(io.type) == filter_type)
        ]

    async def read(self, key: str) -> bool | int | float:
        """Reads a value from a given IO"""
        return await self._api_gateway.read_controller_io(
            cell=self._cell, controller=self._controller_id, io=key
        )

    async def write(self, key: str, value: ValueType):
        """Set a value asynchronously (So a direct read after setting might return still the old value)"""
        await self._ensure_value_type(key, value)

        async with self._io_operation_in_progress:
            await self._api_gateway.write_controller_io(
                cell=self._cell,
                controller=self._controller_id,
                io=key,
                value=value,  # type: ignore
            )

    async def _ensure_value_type(self, key: str, value: ValueType):
        """Checks if the provided value matches the expected type of the IO"""
        io_descriptions = await self.get_io_descriptions()
        io_description = io_descriptions[key]
        io_value_type = IOValueType(io_description.value_type)
        if isinstance(value, bool):
            if io_value_type is not IOValueType.IO_VALUE_DIGITAL:
                raise ValueError(
                    f"Boolean value can only be set at an IO_VALUE_DIGITAL IO and not to {io_value_type}"
                )
        elif isinstance(value, int):
            if io_value_type is not IOValueType.IO_VALUE_ANALOG_INTEGER:
                raise ValueError(
                    f"Integer value can only be set at an IO_VALUE_ANALOG_INTEGER IO and not to {io_value_type}"
                )
        elif isinstance(value, float):
            if io_value_type is not IOValueType.IO_VALUE_ANALOG_FLOATING:
                raise ValueError(
                    f"Float value can only be set at an IO_VALUE_ANALOG_FLOATING IO and not to {io_value_type}"
                )
        else:
            raise ValueError(f"Unexpected type {type(value)}")

    async def wait_for_bool_io(self, io: str, value: bool):
        """Blocks until the requested IO equals the provided value."""
        # TODO proper implementation utilising also the comparison operators
        await self._api_gateway.wait_for_bool_io(
            cell=self._cell, controller=self._controller_id, io=io, value=value
        )



import tempfile
from pathlib import Path
from zipfile import ZipFile

import requests

"""
This script downloads the pre-processed robot models from the wandelbots-js-react-components repository.
These models are already optimized and don't require additional processing.
"""


def get_project_root() -> Path:
    """Get the root directory of the user's project"""
    cwd = Path.cwd()
    markers = ["pyproject.toml", "setup.py", ".git"]

    current = cwd
    while current != current.parent:
        if any((current / marker).exists() for marker in markers):
            return current
        current = current.parent

    return cwd


def get_current_version(models_dir: Path) -> str:
    version_file = models_dir / "version.txt"
    if version_file.exists():
        return version_file.read_text().strip()
    return ""


def get_latest_release_version() -> str:
    api_url = (
        "https://api.github.com/repos/wandelbotsgmbh/wandelbots-js-react-components/releases/latest"
    )
    response = requests.get(api_url)
    if response.status_code != 200:
        raise Exception(f"Failed to get latest release: {response.status_code}")
    return response.json()["tag_name"].lstrip("v")


def download_and_extract(version: str, models_dir: Path) -> None:
    """Download and extract pre-processed models directly to models directory"""
    zip_url = f"https://github.com/wandelbotsgmbh/wandelbots-js-react-components/releases/download/v{version}/no-draco-models.zip"
    print(f"Downloading {zip_url}...")

    response = requests.get(zip_url)
    if response.status_code != 200:
        raise Exception(f"Failed to download: {response.status_code}")

    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_path = Path(tmp_dir)
        zip_path = tmp_path / "models.zip"
        zip_path.write_bytes(response.content)

        with ZipFile(zip_path, "r") as zip_ref:
            zip_ref.extractall(models_dir)


def update_robot_models():
    version = get_latest_release_version()
    models_dir = Path.cwd() / "models"
    current_version = get_current_version(models_dir)

    if version == current_version:
        print("Models are up to date")
        return

    # Create models directory if it doesn't exist
    models_dir.mkdir(parents=True, exist_ok=True)

    # Download and extract pre-processed models
    download_and_extract(version, models_dir)

    # Write version file
    (models_dir / "version.txt").write_text(version)

    print(f"Successfully updated robot models {current_version} ➜ {version}")


if __name__ == "__main__":
    update_robot_models()



from __future__ import annotations

from typing import Annotated, Callable

import pydantic

from nova import api
from nova.actions.io import WriteAction
from nova.actions.motions import CollisionFreeMotion, Motion
from nova.types import MovementControllerFunction, Pose


class ActionLocation(pydantic.BaseModel):
    """A container for an action at a specific path parameter"""

    path_parameter: float = 1.0
    action: WriteAction


# TODO: all actions should be allowed (Action)
ActionContainerItem = Motion | WriteAction | CollisionFreeMotion


class CombinedActions(pydantic.BaseModel):
    """A trajectory of motions and actions"""

    # See: https://docs.pydantic.dev/latest/concepts/serialization/#serialize_as_any-runtime-setting
    items: tuple[
        Annotated[
            pydantic.SerializeAsAny[ActionContainerItem], pydantic.Field(discriminator="type")
        ],
        ...,
    ] = ()

    def __len__(self):
        return len(self.items)

    def __getitem__(self, item):
        return self.items[item]

    def __setattr__(self, key, value):
        if key == "items":
            raise TypeError("Cannot set items directly")
        super().__setattr__(key, value)

    def __iter__(self):
        return iter(self.items)

    def append(self, item: ActionContainerItem):
        super().__setattr__("items", self.items + (item,))

    def _generate_trajectory(
        self,
    ) -> tuple[list[Motion | CollisionFreeMotion], list[ActionLocation]]:
        """Generate two lists: one of Motion objects and another of ActionContainer objects,
        where each ActionContainer wraps a non-Motion action with its path parameter.

        The path parameter is the index of the last Motion object in the list of Motion objects.
        S - M - M - A - A - M - M - A - M - M
        0 - 1 - 2 - 3 - 3 - 3 - 4 - 5 - 5 - 6

        Returns:
            tuple: A tuple containing:
                - list of Motion objects from self.items.
                - list of ActionContainer objects with indexed path parameters.
        """
        motions = []
        actions = []
        last_motion_index = 0

        for item in self.items:
            if isinstance(item, Motion) or isinstance(item, CollisionFreeMotion):
                motions.append(item)
                last_motion_index += 1  # Increment the motion index for each new Motion
            else:
                # Assign the current value of last_motion_index as path_parameter for actions
                actions.append(ActionLocation(path_parameter=last_motion_index, action=item))

        return motions, actions

    @property
    def motions(self) -> list[Motion | CollisionFreeMotion]:
        motions, _ = self._generate_trajectory()
        return motions

    @property
    def actions(self) -> list[ActionLocation]:
        _, actions = self._generate_trajectory()
        return actions

    @property
    def start(self) -> ActionContainerItem | None:
        return self.motions[0] if self.motions else None

    @property
    def end(self) -> ActionContainerItem | None:
        return self.motions[-1] if self.motions else None

    def poses(self) -> list[Pose]:
        """Returns the positions of all motions. If a motion is not a cartesian motion, the position is ignored

        Returns: the positions

        """
        motions, _ = self._generate_trajectory()
        return [
            Pose(position=motion.target.position, orientation=motion.target.orientation)
            for motion in motions
            if isinstance(motion.target, Pose)
        ]

    def positions(self):
        """Returns the positions of all motions. If a motion is not a cartesian motion, the position is ignored

        Returns: the positions

        """
        return [pose.position for pose in self.poses()]

    def orientations(self):
        """Returns the orientations of all motions. If a motion is not a cartesian motion, the orientation is ignored

        Returns: the orientations

        """
        return [pose.orientation for pose in self.poses()]

    def __add__(self, other: CombinedActions) -> CombinedActions:
        return CombinedActions(items=self.items + other.items)

    def to_motion_command(self) -> list[api.models.MotionCommand]:
        motion_commands = []
        for motion in self.motions:
            path = api.models.MotionCommandPath.from_dict(motion.model_dump())
            blending = (
                motion.settings.as_blending_setting()
                if motion.settings.has_blending_settings()
                else None
            )
            limits_override = (
                motion.settings.as_limits_settings()
                if motion.settings.has_limits_override()
                else None
            )
            motion_commands.append(
                api.models.MotionCommand(
                    path=path, blending=blending, limits_override=limits_override
                )
            )
        return motion_commands

    def to_set_io(self) -> list[api.models.SetIO]:
        return [
            api.models.SetIO(
                io=api.models.IOValue(**action.action.model_dump(exclude_unset=True)),
                location=action.path_parameter,
            )
            for action in self.actions
        ]


# TODO: should not be located here
class MovementControllerContext(pydantic.BaseModel):
    combined_actions: CombinedActions
    motion_id: str


MovementController = Callable[[MovementControllerContext], MovementControllerFunction]



from __future__ import annotations

from decouple import config

from nova.cell.cell import Cell
from nova.core.gateway import ApiGateway

LOG_LEVEL = config("LOG_LEVEL", default="INFO")
CELL_NAME = config("CELL_NAME", default="cell")


# TODO: could also extend NovaDevice
class Nova:
    """A high-level Nova client for interacting with robot cells and controllers."""

    _api_client: ApiGateway

    def __init__(
        self,
        *,
        host: str | None = None,
        access_token: str | None = None,
        username: str | None = None,
        password: str | None = None,
        version: str = "v1",
        verify_ssl: bool = True,
    ):
        """
        Initialize the Nova client.

        Args:
            host (str | None): The Nova API host.
            access_token (str | None): An access token for the Nova API.
            username (str | None): Username to authenticate with the Nova API.
            password (str | None): Password to authenticate with the Nova API.
            version (str): The API version to use (default: "v1").
            verify_ssl (bool): Whether or not to verify SSL certificates (default: True).
        """

        self._api_client = ApiGateway(
            host=host,
            access_token=access_token,
            username=username,
            password=password,
            version=version,
            verify_ssl=verify_ssl,
        )

    def cell(self, cell_id: str = CELL_NAME) -> Cell:
        """Returns the cell object with the given ID."""
        return Cell(self._api_client, cell_id)

    async def close(self):
        """Closes the underlying API client session."""
        return await self._api_client.close()

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()
